
from typing import Iterable, Tuple
from .b20161101 import b20161101
from ..constants import *
from ..types import *
from ..io import *

class b20250306(b20161101):
    """
    b20250306 changes the pp datatype from s16 to u16, due
    to mrekk exceeding the integer limit.
    https://osu.ppy.sh/home/changelog/stable40/20250306.2
    https://www.reddit.com/r/osugame/comments/1j3y7fh/
    """
    version = 20250306

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, info.id)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        write_u16(stream, info.stats.pp)
        yield PacketType.BanchoUserStats, stream.data




from typing import Iterable, Tuple
from .b20120818 import b20120818
from ..constants import *
from ..types import *
from ..io import *

class b20121023(b20120818):
    """
    b20121023 adds the "ban/silence info" packet, which locks down account switching
    on the client-side, and tells the user how long they are banned/silenced for.
    """
    version = 20121023
    protocol_version = 11

    @classmethod
    def write_silence_info(cls, length_seconds: int = -1) -> Iterable[Tuple[PacketType, bytes]]:
        # NOTE: If set to "-1", the client will reset the silence
        #       info and allow account switching again.
        stream = MemoryStream()
        write_s32(stream, length_seconds)
        yield PacketType.BanchoSilenceInfo, stream.data




from typing import Iterable, Tuple
from .b20121212 import b20121212
from ..constants import *
from ..types import *
from ..io import *

class b20121221(b20121212):
    """
    b20121221 adds both the image url and redirect url to the title update packet.
    """
    version = 20121221

    @classmethod
    def write_title_update(cls, update: TitleUpdate) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, f"{update.image_url or ''}|{update.redirect_url or ''}")
        yield PacketType.BanchoTitleUpdate, stream.data




from .constants import *
from dataclasses import dataclass, field
from typing import List, Optional
from hashlib import md5

__all__ = [
    "UserInfo",
    "UserPresence",
    "UserStats",
    "UserStatus",
    "UserQuit",
    "Message",
    "Channel",
    "BeatmapInfo",
    "BeatmapInfoReply",
    "BeatmapInfoRequest",
    "ReplayFrame",
    "ScoreFrame",
    "ReplayFrameBundle",
    "MatchSlot",
    "Match",
    "MatchJoin",
    "TitleUpdate"
]

@dataclass
class UserPresence:
    is_irc: bool = False
    timezone: int = 0
    country_index: int = 0
    permissions: Permissions = Permissions.Regular
    longitude: float = 0.0
    latitude: float = 0.0
    city: str = ""

    @property
    def country_name(self) -> str:
        return CountryNames[self.country_index]

    @property
    def country_acronym(self) -> str:
        return CountryAcronyms[self.country_index]

    @property
    def country_string(self) -> str:
        return (
            self.country_name if not self.city else
            f"{self.country_name} / {self.city}"
        )

@dataclass
class UserStats:
    rank: int = 0
    rscore: int = 0
    tscore: int = 0
    accuracy: float = 0.0
    playcount: int = 0
    pp: int = 0

@dataclass
class UserStatus:
    action: Status = Status.Idle
    text: str = ""
    mods: Mods = Mods.NoMod
    mode: Mode = Mode.Osu
    beatmap_checksum: str = ""
    beatmap_id: int = -1
    update_stats: bool = False

    def reset(self) -> None:
        self.action = Status.Idle
        self.text = ""
        self.mods = Mods.NoMod
        self.mode = Mode.Osu
        self.beatmap_checksum = ""
        self.beatmap_id = -1
        self.update_stats = False

@dataclass
class UserInfo:
    id: int = 0
    name: str = ""
    presence: UserPresence = field(default_factory=UserPresence)
    status: UserStatus = field(default_factory=UserStatus)
    stats: UserStats = field(default_factory=UserStats)

    @property
    def avatar_filename(self) -> str:
        return f"{self.id}_000.png"

@dataclass
class UserQuit:
    info: UserInfo = field(default_factory=UserInfo)
    state: QuitState = QuitState.Gone

@dataclass
class Message:
    sender: str
    content: str
    target: str
    sender_id: int = -1

    @property
    def is_direct_message(self) -> bool:
        return not self.target.startswith("#")

@dataclass
class Channel:
    name: str
    topic: str = ""
    owner: str = "BanchoBot"
    user_count: int = 0

@dataclass
class BeatmapInfo:
    index: int
    beatmap_id: int
    beatmapset_id: int
    thread_id: int
    ranked_status: RankedStatus
    checksum: str
    osu_rank: Rank = Rank.N
    taiko_rank: Rank = Rank.N
    fruits_rank: Rank = Rank.N
    mania_rank: Rank = Rank.N

    @property
    def is_ranked(self) -> bool:
        return self.ranked_status in (RankedStatus.Ranked, RankedStatus.Approved)

@dataclass
class BeatmapInfoReply:
    beatmaps: List[BeatmapInfo] = field(default_factory=list)

@dataclass
class BeatmapInfoRequest:
    filenames: List[str] = field(default_factory=list)
    ids: List[int] = field(default_factory=list)

@dataclass
class ReplayFrame:
    button_state: ButtonState = ButtonState.NoButton
    legacy_byte: int = 0
    mouse_x: float = 0.0
    mouse_y: float = 0.0
    time: int = 0

@dataclass
class ScoreFrame:
    time: int
    id: int
    total_300: int
    total_100: int
    total_50: int
    total_geki: int
    total_katu: int
    total_miss: int
    total_score: int
    max_combo: int
    current_combo: int
    perfect: bool
    hp: int
    tag_byte: int
    using_scorev2: bool = False
    combo_portion: float = 0.0
    bonus_portion: float = 0.0

    @property
    def passed(self) -> bool:
        return self.hp != 254

    @property
    def checksum(self) -> str:
        data = (
            f"{self.time}{self.passed}{self.total_300}{self.total_50}{self.total_geki}"
            f"{self.total_katu}{self.total_miss}{self.current_combo}"
            f"{self.max_combo}{self.hp}"
        )
        return md5(data.encode()).hexdigest()

    def total_hits(self, mode: Mode) -> int:
        if mode == Mode.CatchTheBeat:
            return self.total_50 + self.total_100 + self.total_300 + self.total_miss + self.total_katu

        elif mode == Mode.OsuMania:
            return self.total_300 + self.total_100 + self.total_50 + self.total_geki + self.total_katu + self.total_miss

        return self.total_50 + self.total_100 + self.total_300 + self.total_miss

    def accuracy(self, mode: Mode) -> float:
        if self.total_hits(mode) == 0:
            return 0.0

        if mode == Mode.Osu:
            return (
                ((self.total_300 * 300.0) + (self.total_100 * 100.0) + (self.total_50 * 50.0))
                / (self.total_hits(mode) * 300.0)
            )

        elif mode == Mode.Taiko:
            return ((self.total_100 * 0.5) + self.total_300) / self.total_hits(mode)

        elif mode == Mode.CatchTheBeat:
            return (self.total_300 + self.total_100 + self.total_50) / self.total_hits(mode)

        elif mode == Mode.OsuMania:
            return  (
                (
                    (self.total_50 * 50.0) + (self.total_100 * 100.0) + (self.total_katu * 200.0) + ((self.total_300 + self.total_geki) * 300.0)
                )
                / (self.total_hits(mode) * 300.0)
            )

        else:
            return 0.0

    def rank(self, mode: Mode, mods: Mods) -> Rank:
        r300 = self.total_300 / self.total_hits(mode)
        r50 = self.total_50 / self.total_hits(mode)

        if r300 == 1:
            return (
                Rank.XH
                if Mods.Hidden in mods
                or Mods.Flashlight in mods
                else Rank.X
            )

        if r300 > 0.9 and r50 <= 0.01 and self.total_miss == 0:
            return (
                Rank.SH
                if Mods.Hidden in mods
                or Mods.Flashlight in mods
                else Rank.S
            )

        if (r300 > 0.8 and self.total_miss == 0) or (r300 > 0.9):
            return Rank.A

        if (r300 > 0.7 and self.total_miss == 0) or (r300 > 0.8):
            return Rank.B

        if (r300 > 0.6):
            return Rank.C

        return Rank.D

@dataclass
class ReplayFrameBundle:
    action: ReplayAction = ReplayAction.Standard
    frames: List[ReplayFrame] = field(default_factory=list)
    frame: Optional[ScoreFrame] = None
    extra: int = -1
    sequence: Optional[int] = None

@dataclass
class MatchSlot:
    user_id: int = -1
    status: SlotStatus = SlotStatus.Open
    team: SlotTeam = SlotTeam.Neutral
    mods: Mods = Mods.NoMod

    @property
    def has_player(self) -> bool:
        return bool(SlotStatus.HasPlayer & self.status)

    def reset(self) -> None:
        self.user_id = -1
        self.status = SlotStatus.Open
        self.team = SlotTeam.Neutral
        self.mods = Mods.NoMod

@dataclass
class Match:
    id: int = 0
    in_progress: bool = False
    type: MatchType = MatchType.Standard
    mods: Mods = Mods.NoMod
    name: str = ""
    password: str = ""
    beatmap_text: str = ""
    beatmap_id: int = -1
    beatmap_checksum: str = ""
    slots: List[MatchSlot] = field(default_factory=list)
    host_id: int = -1
    mode: Mode = Mode.Osu
    scoring_type: ScoringType = ScoringType.Score
    team_type: TeamType = TeamType.HeadToHead
    freemod: bool = False
    seed: int = 0

@dataclass
class MatchJoin:
    match_id: int = -1
    password: str = ""

@dataclass
class TitleUpdate:
    image_url: str = ""
    redirect_url: str = ""




__author__ = "Lekuru"
__email__ = "contact@lekuru.xyz"
__version__ = "1.1.13"
__license__ = "MIT"

from .utils import select_client, select_latest_client, select_initial_client, resolve_country_index
from .patching import patch, set_protocol_version, set_slot_size
from .chio import BanchoIO
from .io import Stream
from .constants import *
from .types import *




from typing import Tuple, Iterable
from .b296 import b296
from ..constants import *
from ..types import *
from ..io import *

class b298(b296):
    """
    b298 adds a partial implementation of multiplayer, as well as fellow spectators.
    """
    version = 298

    @classmethod
    def write_match_update(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        if match.id > 0xFF:
            # Match IDs greater than 255 are not supported in this client
            return []

        yield PacketType.BanchoMatchUpdate, cls.write_match(match)

    @classmethod
    def write_match_new(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        if match.id > 0xFF:
            # Match IDs greater than 255 are not supported in this client
            return []

        yield PacketType.BanchoMatchNew, cls.write_match(match)

    @classmethod
    def write_match_disband(cls, match_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, match_id)
        yield PacketType.BanchoMatchDisband, stream.data

    @classmethod
    def write_lobby_join(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        yield PacketType.BanchoLobbyJoin, stream.data

    @classmethod
    def write_lobby_part(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        yield PacketType.BanchoLobbyPart, stream.data

    @classmethod
    def write_match_join_success(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchJoinSuccess, cls.write_match(match)
    
    @classmethod
    def write_match_join_fail(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchJoinFail, b""

    @classmethod
    def write_fellow_spectator_joined(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        # Weirdly enough, the client seems to need both of these packets to be sent?
        # If only one is sent, the client will not display the fellow spectator.
        # yield PacketType.BanchoSpectatorJoined, stream.data
        yield PacketType.BanchoFellowSpectatorJoined, stream.data

    @classmethod
    def write_fellow_spectator_left(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        yield PacketType.BanchoFellowSpectatorLeft, stream.data

    @classmethod
    def read_lobby_join(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_lobby_part(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_create(cls, stream: MemoryStream) -> Match:
        return cls.read_match(stream)

    @classmethod
    def read_match_join(cls, stream: MemoryStream) -> MatchJoin:
        return MatchJoin(read_s32(stream))

    @classmethod
    def read_match_part(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_change_slot(cls, stream: MemoryStream) -> int:
        return read_s32(stream)
    
    @classmethod
    def read_match_ready(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_lock(cls, stream: MemoryStream) -> int:
        return read_s32(stream)

    @classmethod
    def read_match_change_settings(cls, stream: MemoryStream) -> Match:
        return cls.read_match(stream)

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        slots_open = [
            slot.status == SlotStatus.Open
            for slot in match.slots
        ]

        slots_used = [
            slot.has_player
            for slot in match.slots
        ]

        slots_ready = [
            slot.status == SlotStatus.Ready
            for slot in match.slots
        ]

        stream = MemoryStream()
        write_u8(stream, match.id)
        write_u8(stream, match.type)
        write_string(stream, match.name)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)
        write_bool_list(stream, slots_open)
        write_bool_list(stream, slots_used)
        write_bool_list(stream, slots_ready)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        return stream.data
    
    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u8(stream)
        match.type = MatchType(read_u8(stream))
        match.name = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)

        slots_open = read_bool_list(stream)
        slots_used = read_bool_list(stream)
        slots_ready = read_bool_list(stream)
        match.slots = []

        for i in range(cls.slot_size):
            slot = MatchSlot()
            slot.status = SlotStatus.Open if slots_open[i] else SlotStatus.Locked
            slot.status = SlotStatus.NotReady if slots_used[i] else slot.status
            slot.status = SlotStatus.Ready if slots_ready[i] else slot.status

            if slot.has_player:
                slot.user_id = read_s32(stream)

            match.slots.append(slot)

        return match




from typing import Tuple, Iterable
from .b298 import b298
from ..constants import *
from ..types import *
from ..io import *

class b312(b298):
    """
    b312 adds the match start & update packets, as well
    as the "InProgress" boolean inside the match struct.
    """
    version = 312

    @classmethod
    def write_match_start(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchStart, b""

    @classmethod
    def write_match_score_update(cls, frame: ScoreFrame) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        cls.write_score_frame(stream, frame)
        yield PacketType.BanchoMatchScoreUpdate, stream.data

    @classmethod
    def read_match_start(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_score_update(cls, stream: MemoryStream) -> ScoreFrame:
        return cls.read_score_frame(stream)
    
    @classmethod
    def read_match_complete(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        slots_open = [
            slot.status == SlotStatus.Open
            for slot in match.slots
        ]

        slots_used = [
            slot.has_player
            for slot in match.slots
        ]

        slots_ready = [
            slot.status == SlotStatus.Ready
            for slot in match.slots
        ]

        stream = MemoryStream()
        write_u8(stream, match.id)
        write_boolean(stream, match.in_progress)
        write_u8(stream, match.type)
        write_string(stream, match.name)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)
        write_bool_list(stream, slots_open)
        write_bool_list(stream, slots_used)
        write_bool_list(stream, slots_ready)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        return stream.data

    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u8(stream)
        match.in_progress = read_boolean(stream)
        match.type = MatchType(read_u8(stream))
        match.name = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)

        slots_open = read_bool_list(stream)
        slots_used = read_bool_list(stream)
        slots_ready = read_bool_list(stream)
        match.slots = []

        for i in range(cls.slot_size):
            slot = MatchSlot()
            slot.status = SlotStatus.Open if slots_open[i] else SlotStatus.Locked
            slot.status = SlotStatus.NotReady if slots_used[i] else slot.status
            slot.status = SlotStatus.Ready if slots_ready[i] else slot.status

            if slot.has_player:
                slot.user_id = read_s32(stream)

            match.slots.append(slot)

        return match




from typing import Iterable, Tuple
from .b20121203 import b20121203
from ..constants import *
from ..types import *
from ..io import *

class b20121207(b20121203):
    """
    b20121207 allows tourney clients to request multiplayer match information.
    """
    version = 20121207

    @classmethod
    def read_tournament_match_info(cls, stream: MemoryStream) -> int:
        return read_s32(stream)




from typing import Iterable, Tuple
from .b20130131 import b20130131
from ..constants import *
from ..types import *
from ..io import *

class b20130209(b20130131):
    """
    b20130209 adds the target silenced packet, which
    informs the client, that a user has been silenced.
    """
    version = 20130209

    @classmethod
    def write_target_is_silenced(cls, username: str) -> Iterable[Tuple[PacketType, bytes]]:
        _, data = next(cls.write_message(Message("", "", username)))
        yield PacketType.BanchoTargetIsSilenced, data




from typing import Iterable, Tuple
from .b20151107 import b20151107
from ..constants import *
from ..types import *
from ..io import *

class b20160404(b20151107):
    """
    b20160404 adds a sequence number to the replay frame
    packet, to allow for replaying frames.
    """
    version = 20160404

    @classmethod
    def read_spectate_frames(cls, stream: MemoryStream) -> ReplayFrameBundle:
        if cls.protocol_version >= 18:
            extra = read_u32(stream)

        frames = [
            cls.read_replay_frame(stream)
            for _ in range(read_u16(stream))
        ]
        action = ReplayAction(read_u8(stream))
        frame = None

        if stream.available() > 2:
            frame = cls.read_score_frame(stream)

        sequence = read_u16(stream)
        return ReplayFrameBundle(action, frames, frame, extra, sequence)

    @classmethod
    def write_spectate_frames(cls, bundle: ReplayFrameBundle) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if cls.protocol_version >= 18:
            write_s32(stream, bundle.extra)

        write_u16(stream, len(bundle.frames))

        for frame in bundle.frames:
            stream.write(cls.write_replay_frame(frame))

        write_u8(stream, bundle.action)

        if bundle.frame:
            cls.write_score_frame(stream, bundle.frame)

        write_u16(stream, bundle.sequence or 0)
        yield PacketType.BanchoSpectateFrames, stream.data




from typing import Iterable, Tuple
from .b340 import b340
from ..constants import *
from ..types import *
from ..io import *

class b342(b340):
    """
    b342 adds the multiplayer skip packets.
    """
    version = 342

    @classmethod
    def write_match_skip(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchSkip, b''

    @classmethod
    def read_match_skip_request(cls, stream: MemoryStream) -> None:
        pass




from typing import Iterable, Tuple
from .b349 import b349
from ..constants import *
from ..types import *
from ..io import *

class b354(b349):
    """
    b354 adds the beatmap info packets.
    """
    version = 354

    @classmethod
    def read_beatmap_info_request(cls, stream: MemoryStream) -> BeatmapInfoRequest:
        return BeatmapInfoRequest([read_string(stream) for _ in range(read_u32(stream))])

    @classmethod
    def write_beatmap_info_reply(cls, reply: BeatmapInfoReply) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, len(reply.beatmaps))

        for info in reply.beatmaps:
            write_s16(stream, info.index)
            write_s32(stream, info.beatmap_id)
            write_s32(stream, info.beatmapset_id)
            write_s32(stream, info.thread_id)
            write_boolean(stream, info.is_ranked)
            write_s8(stream, info.osu_rank)
            write_string(stream, info.checksum)

        yield PacketType.BanchoBeatmapInfoReply, stream.data




from typing import Iterable, Tuple
from .b20140731 import b20140731
from ..constants import *
from ..types import *
from ..io import *

class b20141104(b20140731):
    """
    b20141104 allows the client to switch to a tournament server & abort a match.
    """
    version = 20141104

    @classmethod
    def write_switch_tournament_server(cls, server: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, server)
        yield PacketType.BanchoSwitchTournamentServer, stream.data

    @classmethod
    def write_match_abort(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchAbort, b''




from typing import Any, Tuple, Iterable
from .io import Stream, MemoryStream, AsyncStream
from .constants import PacketType

class BanchoIO:
    """
    BanchoIO is an interface that wraps the basic methods for
    reading and writing packets to a Bancho client.
    """
    version: int = 0
    slot_size: int = 8
    header_size: int = 6
    protocol_version: int = 0
    disable_compression: bool = False
    requires_status_updates: bool = True

    @classmethod
    def read_packet(cls, stream: Stream) -> Tuple[PacketType, Any]:
        """
        Reads a packet from the stream, and returns the packet type and decoded data.
        The type of the decoded data depends on the received packet.
        """
        ...

    @classmethod
    def write_packet(cls, stream: Stream, packet: PacketType, *args) -> None:
        """
        Encodes a packet and writes it to the stream.
        """
        ...

    @classmethod
    async def read_packet_async(cls, stream: AsyncStream) -> Tuple[PacketType, Any]:
        """
        Reads a packet from the stream asynchronously, and returns the packet type and
        decoded data. The type of the decoded data depends on the received packet.
        """
        ...

    @classmethod
    async def write_packet_async(cls, stream: AsyncStream, packet: PacketType, *args) -> None:
        """
        Encodes a packet and writes it to the stream, asynchronously.
        """
        ...

    @classmethod
    def implements_packet(cls, packet: PacketType) -> bool:
        """
        Returns whether the current client version implements the given packet.
        """
        return getattr(cls, packet.handler_name, None) is not None

    @classmethod
    def read_packet_from_bytes(cls, data: bytes) -> Tuple[PacketType, Any]:
        """
        Reads a packet from the given bytes, and returns the packet type and decoded data.
        """
        stream = MemoryStream(data)
        return cls.read_packet(stream)

    @classmethod
    def read_many_packets_from_bytes(cls, data: bytes) -> Iterable[Tuple[PacketType, Any]]:
        """
        Reads multiple packets from the given bytes, and yields the packet type and decoded data.
        """
        stream = MemoryStream(data)

        while stream.available() >= cls.header_size:
            yield cls.read_packet(stream)

    @classmethod
    def write_packet_to_bytes(cls, packet: PacketType, *args) -> bytes:
        """
        Encodes a packet and returns it as bytes.
        """
        stream = MemoryStream()
        cls.write_packet(stream, packet, *args)
        return stream.data

    @classmethod
    def write_many_packets_to_bytes(cls, packets: Iterable[Tuple[PacketType, Any]]) -> bytes:
        """
        Encodes multiple packets and returns them as bytes.
        """
        stream = MemoryStream()

        for packet, *args in packets:
            cls.write_packet(stream, packet, *args)

        return stream.data




from typing import Iterable, Tuple
from .b591 import b591
from ..constants import *
from ..types import *
from ..io import *

class b613(b591):
    """
    b613 allows the client to leave channels.
    """
    version = 613

    @classmethod
    def read_channel_leave(cls, stream: MemoryStream) -> str:
        return read_string(stream)




from typing import Iterable, Tuple
from .b1796 import b1796
from ..constants import *
from ..types import *
from ..io import *

class b1797(b1796):
    """
    b1797 adds the rank field to re-enable user rank sorting.
    """
    version = 1797
    protocol_version = 7

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, cls.convert_user_id(info))
        write_string(stream, info.name)
        write_u8(stream, AvatarExtension.Png)
        write_u8(stream, info.presence.timezone+24)
        write_u8(stream, info.presence.country_index)
        write_string(stream, info.presence.city)
        write_u8(stream, info.presence.permissions)
        write_f32(stream, info.presence.longitude)
        write_f32(stream, info.presence.latitude)

        if cls.protocol_version >= 7:
            write_s32(stream, info.stats.rank)

        yield PacketType.BanchoUserPresence, stream.data




from typing import Iterable, Tuple, Union
from .b20160404 import b20160404
from ..constants import *
from ..types import *
from ..io import *

class b20161101(b20160404):
    """
    b20161101 introduced the loved status to the beatmap info reply.
    """
    version = 20161101

    @classmethod
    def convert_ranked_status(cls, status: Union[RankedStatus, int]) -> int:
        return status




from typing import Iterable, Tuple
from .b504 import b504
from ..constants import *
from ..types import *
from ..io import *

class b535(b504):
    """
    b535 adds the scoring & team types for multiplayer matches, as well
    as the tag multiplayer game mode.
    """
    version = 535
    protocol_version = 3

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        stream = MemoryStream()
        write_u8(stream, match.id)
        write_boolean(stream, match.in_progress)
        write_u8(stream, match.type)
        write_u16(stream, match.mods.value)
        write_string(stream, match.name)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)

        for slot in match.slots:
            write_u8(stream, slot.status.value)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        write_s32(stream, match.host_id)
        write_u8(stream, match.mode)

        if cls.protocol_version >= 3:
            write_u8(stream, match.scoring_type)
            write_u8(stream, match.team_type)

        return stream.data

    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u8(stream)
        match.in_progress = read_boolean(stream)
        match.type = MatchType(read_u8(stream))
        match.mods = Mods(read_u16(stream))
        match.name = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)
        match.slots = [
            MatchSlot(status=SlotStatus(read_u8(stream)))
            for _ in range(cls.slot_size)
        ]

        for slot in match.slots:
            if slot.has_player:
                slot.user_id = read_s32(stream)

        match.host_id = read_s32(stream)
        match.mode = Mode(read_u8(stream))

        if cls.protocol_version < 3:
            return match

        # There are multiple versions of b535, which have different
        # levels of implementation for the new match data, so we need
        # to check if the data is available before reading it.

        if stream.available() > 0:
            match.scoring_type = ScoringType(read_u8(stream))

        if stream.available() > 0:
            match.team_type = TeamType(read_u8(stream))

        return match

    @classmethod
    def write_score_frame(cls, stream: MemoryStream, frame: ScoreFrame) -> None:
        write_s32(stream, frame.time)
        write_u8(stream, frame.id)
        write_u16(stream, frame.total_300)
        write_u16(stream, frame.total_100)
        write_u16(stream, frame.total_50)
        write_u16(stream, frame.total_geki)
        write_u16(stream, frame.total_katu)
        write_u16(stream, frame.total_miss)
        write_u32(stream, frame.total_score)
        write_u16(stream, frame.max_combo)
        write_u16(stream, frame.current_combo)
        write_boolean(stream, frame.perfect)
        write_u8(stream, frame.hp)
        write_u8(stream, frame.tag_byte)

    @classmethod
    def read_score_frame(cls, stream: MemoryStream) -> ScoreFrame:
        frame = ScoreFrame(
            time=read_s32(stream),
            id=read_u8(stream),
            total_300=read_u16(stream),
            total_100=read_u16(stream),
            total_50=read_u16(stream),
            total_geki=read_u16(stream),
            total_katu=read_u16(stream),
            total_miss=read_u16(stream),
            total_score=read_u32(stream),
            max_combo=read_u16(stream),
            current_combo=read_u16(stream),
            perfect=read_boolean(stream),
            hp=read_u8(stream),
            tag_byte=0
        )

        if stream.available() > 0:
            # Similar to the scoring type in the match data; the
            # client has a new implementation for the tag byte
            frame.tag_byte = read_u8(stream)

        return frame
    
    @classmethod
    def write_title_update(cls, update: TitleUpdate) -> Iterable[Tuple[PacketType, bytes]]:
        # This is currently only used to refresh the title image, without sending the link
        # over the packet, i.e. it's still using `/web/osu-title-image.php` under the hood.
        yield PacketType.BanchoTitleUpdate, b""




from typing import Iterable, Tuple
from .b1800 import b1800
from ..constants import *
from ..types import *
from ..io import *

class b1817(b1800):
    """
    b1817 adds the ability to send multiplayer match invites to other users.
    """
    version = 1817

    @classmethod
    def write_invite(cls, invite_message: Message) -> Iterable[Tuple[PacketType, bytes]]:
        _, data = next(cls.write_message(invite_message))
        yield PacketType.BanchoInvite, data

    @classmethod
    def read_invite(cls, stream: MemoryStream) -> int:
        return read_s32(stream)




from typing import Iterable, Tuple
from .b20130118 import b20130118
from ..constants import *
from ..types import *
from ..io import *

class b20130131(b20130118):
    """
    b20130121 allows users to block private messages from non-friends.
    """
    version = 20130131

    @classmethod
    def write_user_dms_blocked(cls, username: str) -> Iterable[Tuple[PacketType, bytes]]:
        _, data = next(cls.write_message(Message("", "", username)))
        yield PacketType.BanchoUserDmsBlocked, data

    @classmethod
    def read_change_friend_only_dms(cls, stream: MemoryStream) -> bool:
        return read_s32(stream) == 1




from typing import Iterable, Tuple
from .b320 import b320
from ..constants import *
from ..types import *
from ..io import *

class b323(b320):
    """
    b323 changes the structure of user stats
    and adds the "MatchChangeBeatmap" packet
    """
    version = 323

    @classmethod
    def convert_input_packet(cls, packet: int) -> PacketType:
        if packet == 11:
            # "IrcJoin" packet
            return PacketType.BanchoIrcJoin
        
        if packet == 50:
            # "MatchChangeBeatmap" packet
            return PacketType.OsuMatchChangeBeatmap

        if packet > 11 and packet <= 45:
            packet -= 1

        return PacketType(packet)

    @classmethod
    def convert_output_packet(cls, packet: PacketType) -> int:
        if packet is PacketType.BanchoIrcJoin:
            # "IrcJoin" packet
            return 11

        if packet is PacketType.OsuMatchChangeBeatmap:
            # "MatchChangeBeatmap" packet
            return 50

        if packet >= 11 and packet < 45:
            return packet.value + 1

        return packet.value

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_stats = info.status.update_stats

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_boolean(stream, write_stats)

        if write_stats:
            write_string(stream, info.name)
            write_u64(stream, info.stats.rscore)
            write_f32(stream, info.stats.accuracy)
            write_u32(stream, info.stats.playcount)
            write_u64(stream, info.stats.tscore)
            write_u32(stream, info.stats.rank)
            write_string(stream, info.avatar_filename)
            write_u8(stream, info.presence.timezone+24)
            write_string(stream, info.presence.country_string)

        stream.write(cls.write_status_update(info.status))
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        # We assume that the client has not seen this user before, so
        # we send two packets: one for the user stats, and one for the "presence".
        info.status.update_stats = True
        yield next(cls.write_user_stats(info))

        info.status.update_stats = False
        yield next(cls.write_user_stats(info))

    @classmethod
    def read_match_change_beatmap(cls, stream: MemoryStream) -> Match:
        return cls.read_match(stream)




from .b1797 import b1797
from ..constants import *
from ..types import *
from ..io import *

class b1800(b1797):
    """
    b1800 deprecates the usage of gzip compression inside packets.
    """
    requires_status_updates = False
    version = 1800

    @classmethod
    def write_packet(cls, stream: Stream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            compression_enabled = False
            write_u16(output_stream, packet_id)
            write_boolean(output_stream, compression_enabled)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            stream.write(output_stream.data)
            output_stream.clear()

    @classmethod
    async def write_packet_async(cls, stream: AsyncStream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            compression_enabled = False
            write_u16(output_stream, packet_id)
            write_boolean(output_stream, compression_enabled)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            await stream.write(output_stream.data)
            output_stream.clear()




from functools import cached_property
from enum import IntFlag, IntEnum
from re import compile

__all__ = [
    "PacketType",
    "Status",
    "Mode",
    "LoginError",
    "Permissions",
    "QuitState",
    "AvatarExtension",
    "PresenceFilter",
    "Completeness",
    "ReplayAction",
    "ButtonState",
    "Rank",
    "Mods",
    "MatchType",
    "ScoringType",
    "TeamType",
    "SlotStatus",
    "SlotTeam",
    "RankedStatus",
    "InactiveAccountMessage",
    "Countries",
    "CountryNames",
    "CountryAcronyms"
]

# Regex to convert camelCase to snake_case
# Example: "OsuSendUserStatus" -> "osu_send_user_status"
convert_pattern = compile(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])")

class PacketType(IntEnum):
    OsuUserStatus                  = 0
    OsuMessage                     = 1
    OsuExit                        = 2
    OsuStatusUpdateRequest         = 3
    OsuPong                        = 4
    BanchoLoginReply               = 5
    BanchoCommandError             = 6
    BanchoMessage                  = 7
    BanchoPing                     = 8
    BanchoIrcChangeUsername        = 9
    BanchoIrcQuit                  = 10
    BanchoUserStats                = 11
    BanchoUserQuit                 = 12
    BanchoSpectatorJoined          = 13
    BanchoSpectatorLeft            = 14
    BanchoSpectateFrames           = 15
    OsuStartSpectating             = 16
    OsuStopSpectating              = 17
    OsuSpectateFrames              = 18
    BanchoVersionUpdate            = 19
    OsuErrorReport                 = 20
    OsuCantSpectate                = 21
    BanchoSpectatorCantSpectate    = 22
    BanchoGetAttention             = 23
    BanchoAnnounce                 = 24
    OsuPrivateMessage              = 25
    BanchoMatchUpdate              = 26
    BanchoMatchNew                 = 27
    BanchoMatchDisband             = 28
    OsuLobbyPart                   = 29
    OsuLobbyJoin                   = 30
    OsuMatchCreate                 = 31
    OsuMatchJoin                   = 32
    OsuMatchPart                   = 33
    BanchoLobbyJoin                = 34
    BanchoLobbyPart                = 35
    BanchoMatchJoinSuccess         = 36
    BanchoMatchJoinFail            = 37
    OsuMatchChangeSlot             = 38
    OsuMatchReady                  = 39
    OsuMatchLock                   = 40
    OsuMatchChangeSettings         = 41
    BanchoFellowSpectatorJoined    = 42
    BanchoFellowSpectatorLeft      = 43
    OsuMatchStart                  = 44
    BanchoMatchStart               = 46
    OsuMatchScoreUpdate            = 47
    BanchoMatchScoreUpdate         = 48
    OsuMatchComplete               = 49
    BanchoMatchTransferHost        = 50
    OsuMatchChangeMods             = 51
    OsuMatchLoadComplete           = 52
    BanchoMatchAllPlayersLoaded    = 53
    OsuMatchNoBeatmap              = 54
    OsuMatchNotReady               = 55
    OsuMatchFailed                 = 56
    BanchoMatchPlayerFailed        = 57
    BanchoMatchComplete            = 58
    OsuMatchHasBeatmap             = 59
    OsuMatchSkipRequest            = 60
    BanchoMatchSkip                = 61
    BanchoUnauthorized             = 62
    OsuChannelJoin                 = 63
    BanchoChannelJoinSuccess       = 64
    BanchoChannelAvailable         = 65
    BanchoChannelRevoked           = 66
    BanchoChannelAvailableAutojoin = 67
    OsuBeatmapInfoRequest          = 68
    BanchoBeatmapInfoReply         = 69
    OsuMatchTransferHost           = 70
    BanchoLoginPermissions         = 71
    BanchoFriendsList              = 72
    OsuFriendsAdd                  = 73
    OsuFriendsRemove               = 74
    BanchoProtocolNegotiation      = 75
    BanchoTitleUpdate              = 76
    OsuMatchChangeTeam             = 77
    OsuChannelLeave                = 78
    OsuReceiveUpdates              = 79
    BanchoMonitor                  = 80
    BanchoMatchPlayerSkipped       = 81
    OsuSetIrcAwayMessage           = 82
    BanchoUserPresence             = 83
    OsuUserStatsRequest            = 85
    BanchoRestart                  = 86
    OsuInvite                      = 87
    BanchoInvite                   = 88
    BanchoChannelInfoComplete      = 89
    OsuMatchChangePassword         = 90
    BanchoMatchChangePassword      = 91
    BanchoSilenceInfo              = 92
    OsuTournamentMatchInfo         = 93
    BanchoUserSilenced             = 94
    BanchoUserPresenceSingle       = 95
    BanchoUserPresenceBundle       = 96
    OsuPresenceRequest             = 97
    OsuPresenceRequestAll          = 98
    OsuChangeFriendOnlyDms         = 99
    BanchoUserDmsBlocked           = 100
    BanchoTargetIsSilenced         = 101
    BanchoVersionUpdateForced      = 102
    BanchoSwitchServer             = 103
    BanchoAccountRestricted        = 104
    BanchoRTX                      = 105
    BanchoMatchAbort               = 106
    BanchoSwitchTournamentServer   = 107
    OsuTournamentJoinMatchChannel  = 108
    OsuTournamentLeaveMatchChannel = 109

    # Packets that are unused today, but used in legacy clients
    BanchoIrcJoin         = 0xFFFF
    OsuMatchChangeBeatmap = 0xFFFE

    @cached_property
    def is_server_packet(self) -> bool:
        return self.name.startswith("Bancho")

    @cached_property
    def is_client_packet(self) -> bool:
        return self.name.startswith("Osu")

    @cached_property
    def handler_name(self) -> str:
        name = convert_pattern.sub("_", self.name).lower()
        name = name.replace("osu_", "read_")
        name = name.replace("bancho_", "write_")
        return name

class Status(IntEnum):
    Idle         = 0
    Afk          = 1
    Playing      = 2
    Editing      = 3
    Modding      = 4
    Multiplayer  = 5
    Watching     = 6
    Unknown      = 7
    Testing      = 8
    Submitting   = 9
    Paused       = 10
    Lobby        = 11
    Multiplaying = 12
    OsuDirect    = 13

    # Unused in later versions, but required for compatibility
    StatsUpdate = 10

class Mode(IntEnum):
    Osu          = 0
    Taiko        = 1
    CatchTheBeat = 2
    OsuMania     = 3

    @classmethod
    def from_alias(cls, input: str):
        mapping = {
            'std': Mode.Osu,
            'osu': Mode.Osu,
            'taiko': Mode.Taiko,
            'fruits': Mode.CatchTheBeat,
            'ctb': Mode.CatchTheBeat,
            'catch': Mode.CatchTheBeat,
            'mania': Mode.OsuMania
        }

        if input not in mapping:
            return

        return mapping[input]

    @property
    def formatted(self) -> str:
        return {
            Mode.Osu: 'osu!',
            Mode.Taiko: 'Taiko',
            Mode.CatchTheBeat: 'CatchTheBeat',
            Mode.OsuMania: 'osu!mania'
        }[self]

    @property
    def alias(self) -> str:
        return {
            Mode.Osu: 'osu',
            Mode.Taiko: 'taiko',
            Mode.CatchTheBeat: 'fruits',
            Mode.OsuMania: 'mania'
        }[self]

class LoginError(IntEnum):
    InvalidLogin          = -1
    InvalidVersion        = -2
    UserBanned            = -3
    UserInactive          = -4
    ServerError           = -5
    UnauthorizedTestBuild = -6

class Permissions(IntFlag):
    NoPermissions = 0
    Regular       = 1 << 0
    BAT           = 1 << 1
    Supporter     = 1 << 2
    Friend        = 1 << 3
    Peppy         = 1 << 4
    Tournament    = 1 << 5

class QuitState(IntEnum):
    Gone         = 0
    OsuRemaining = 1
    IrcRemaining = 2

class AvatarExtension(IntEnum):
    Empty = 0
    Png   = 1
    Jpg   = 2

class PresenceFilter(IntEnum):
    NoPlayers = 0
    All       = 1
    Friends   = 2

class Completeness(IntEnum):
    StatusOnly = 0
    Statistics = 1
    Full       = 2

class ReplayAction(IntEnum):
    Standard      = 0
    NewSong       = 1
    Skip          = 2
    Completion    = 3
    Fail          = 4
    Pause         = 5
    Unpause       = 6
    SongSelect    = 7
    WatchingOther = 8

class ButtonState(IntFlag):
    NoButton = 0
    Left1    = 1 << 0
    Right1   = 1 << 1
    Left2    = 1 << 2
    Right2   = 1 << 3
    Smoke    = 1 << 4

class Rank(IntEnum):
    XH = 0
    SH = 1
    X  = 2
    S  = 3
    A  = 4
    B  = 5
    C  = 6
    D  = 7
    F  = 8
    N  = 9

class Mods(IntFlag):
    NoMod       = 0
    NoFail      = 1 << 0
    Easy        = 1 << 1
    NoVideo     = 1 << 2 # replaced by "Touchscreen" in later versions
    Hidden      = 1 << 3
    HardRock    = 1 << 4
    SuddenDeath = 1 << 5
    DoubleTime  = 1 << 6
    Relax       = 1 << 7
    HalfTime    = 1 << 8
    Nightcore   = 1 << 9 # used as "Taiko" mod in older versions
    Flashlight  = 1 << 10
    Autoplay    = 1 << 11
    SpunOut     = 1 << 12
    Autopilot   = 1 << 13
    Perfect     = 1 << 14
    Key4        = 1 << 15
    Key5        = 1 << 16
    Key6        = 1 << 17
    Key7        = 1 << 18
    Key8        = 1 << 19
    FadeIn      = 1 << 20
    Random      = 1 << 21
    Cinema      = 1 << 22
    Target      = 1 << 23
    Key9        = 1 << 24
    KeyCoop     = 1 << 25
    Key1        = 1 << 26
    Key3        = 1 << 27
    Key2        = 1 << 28
    ScoreV2     = 1 << 29
    Mirror      = 1 << 30

    KeyMod = Key1 | Key2 | Key3 | Key4 | Key5 | Key6 | Key7 | Key8 | Key9 | KeyCoop
    FreeModAllowed = NoFail | Easy | Hidden | HardRock | SuddenDeath | Flashlight | FadeIn | Relax | Autopilot | SpunOut | KeyMod
    SpeedMods = DoubleTime | HalfTime| Nightcore

    @property
    def members(self) -> list:
        return [flag for flag in Mods if self & flag]

    @property
    def short(self) -> str:
        if not self:
            return "NM"

        return "".join(
            [ModAcronyms.get(flag, "") for flag in self.members]
        )
    
    @classmethod
    def from_string(cls, acronym_string: str) -> "Mods":
        mods = Mods.NoMod

        if not acronym_string:
            return mods

        # Parse mods into multiple acronyms
        # .e.g. "HDHR" -> ["HR", "HD"]
        parsed_mods = [
            acronym_string[idx : idx + 2].upper()
            for idx in range(0, len(acronym_string), 2)
        ]

        for acronym in parsed_mods:
            mod = ModAcronymsFromString.get(acronym)

            if mod:
                mods |= mod

        return mods

class MatchType(IntEnum):
    Standard  = 0
    Powerplay = 1

class ScoringType(IntEnum):
    Score    = 0
    Accuracy = 1
    Combo    = 2
    ScoreV2  = 3

class TeamType(IntEnum):
    HeadToHead = 0
    TagCoop    = 1
    TeamVs     = 2
    TagTeamVs  = 3

class SlotStatus(IntFlag):
    Open      = 1 << 0
    Locked    = 1 << 1
    NotReady  = 1 << 2
    Ready     = 1 << 3
    NoMap     = 1 << 4
    Playing   = 1 << 5
    Complete  = 1 << 6
    Quit      = 1 << 7
    HasPlayer = NotReady | Ready | NoMap | Playing | Complete

class SlotTeam(IntEnum):
    Neutral = 0
    Blue    = 1
    Red     = 2

    @property
    def opposite(self) -> "SlotTeam":
        return SlotTeam.Red if self != SlotTeam.Red else SlotTeam.Blue

class RankedStatus(IntEnum):
    NotSubmitted = -1
    Pending      = 0
    Ranked       = 1
    Approved     = 2
    Qualified    = 3
    Loved        = 4

InactiveAccountMessage = (
    "Your account is not yet activated. "
    "Please check your email for activation instructions!"
)

ModAcronyms = {
    Mods.NoMod: "NM",
    Mods.NoFail: "NF",
    Mods.Easy: "EZ",
    Mods.NoVideo: "NV",
    Mods.Hidden: "HD",
    Mods.HardRock: "HR",
    Mods.SuddenDeath: "SD",
    Mods.DoubleTime: "DT",
    Mods.Relax: "RX",
    Mods.HalfTime: "HT",
    Mods.Nightcore: "NC",
    Mods.Flashlight: "FL",
    Mods.Autoplay: "AT",
    Mods.SpunOut: "SO",
    Mods.Autopilot: "AP",
    Mods.Perfect: "PF",
    Mods.Key4: "K4",
    Mods.Key5: "K5",
    Mods.Key6: "K6",
    Mods.Key7: "K7",
    Mods.Key8: "K8",
    Mods.FadeIn: "FI",
    Mods.Random: "RD",
    Mods.Cinema: "CN",
    Mods.Target: "TP",
    Mods.Key9: "K9",
    Mods.KeyCoop: "CP",
    Mods.Key1: "K1",
    Mods.Key3: "K3",
    Mods.Key2: "K2",
    Mods.ScoreV2: "V2",
    Mods.Mirror: "MR",
    Mods.SpeedMods: "",
    Mods.KeyMod: "",
    Mods.FreeModAllowed: ""
}

ModAcronymsFromString = {
    "NM": Mods.NoMod,
    "NF": Mods.NoFail,
    "EZ": Mods.Easy,
    "NV": Mods.NoVideo,
    "HD": Mods.Hidden,
    "HR": Mods.HardRock,
    "SD": Mods.SuddenDeath,
    "DT": Mods.DoubleTime,
    "RX": Mods.Relax,
    "HT": Mods.HalfTime,
    "NC": Mods.Nightcore,
    "FL": Mods.Flashlight,
    "AT": Mods.Autoplay,
    "SO": Mods.SpunOut,
    "AP": Mods.Autopilot,
    "PF": Mods.Perfect,
    "K4": Mods.Key4,
    "K5": Mods.Key5,
    "K6": Mods.Key6,
    "K7": Mods.Key7,
    "K8": Mods.Key8,
    "FI": Mods.FadeIn,
    "RD": Mods.Random,
    "CN": Mods.Cinema,
    "TP": Mods.Target,
    "K9": Mods.Key9,
    "CP": Mods.KeyCoop,
    "K1": Mods.Key1,
    "K3": Mods.Key3,
    "K2": Mods.Key2,
    "V2": Mods.ScoreV2,
    "MR": Mods.Mirror
}

Countries = {
    "XX": "Unknown",
    "OC": "Oceania",
    "EU": "Europe",
    "AD": "Andorra",
    "AE": "United Arab Emirates",
    "AF": "Afghanistan",
    "AG": "Antigua and Barbuda",
    "AI": "Anguilla",
    "AL": "Albania",
    "AM": "Armenia",
    "AN": "Netherlands Antilles",
    "AO": "Angola",
    "AQ": "Antarctica",
    "AR": "Argentina",
    "AS": "American Samoa",
    "AT": "Austria",
    "AU": "Australia",
    "AW": "Aruba",
    "AZ": "Azerbaijan",
    "BA": "Bosnia and Herzegovina",
    "BB": "Barbados",
    "BD": "Bangladesh",
    "BE": "Belgium",
    "BF": "Burkina Faso",
    "BG": "Bulgaria",
    "BH": "Bahrain",
    "BI": "Burundi",
    "BJ": "Benin",
    "BM": "Bermuda",
    "BN": "Brunei Darussalam",
    "BO": "Bolivia",
    "BR": "Brazil",
    "BS": "The Bahamas",
    "BT": "Bhutan",
    "BV": "Bouvet Island",
    "BW": "Botswana",
    "BY": "Belarus",
    "BZ": "Belize",
    "CA": "Canada",
    "CC": "Cocos (Keeling) Islands",
    "CD": "Democratic Republic of the Congo",
    "CF": "Central African Republic",
    "CG": "Republic of the Congo",
    "CH": "Switzerland",
    "CI": "Côte d'Ivoire",
    "CK": "Cook Islands",
    "CL": "Chile",
    "CM": "Cameroon",
    "CN": "China",
    "CO": "Colombia",
    "CR": "Costa Rica",
    "CU": "Cuba",
    "CV": "Cape Verde",
    "CX": "Christmas Island",
    "CY": "Cyprus",
    "CZ": "Czech Republic",
    "DE": "Germany",
    "DJ": "Djibouti",
    "DK": "Denmark",
    "DM": "Dominica",
    "DO": "Dominican Republic",
    "DZ": "Algeria",
    "EC": "Ecuador",
    "EE": "Estonia",
    "EG": "Egypt",
    "EH": "Western Sahara",
    "ER": "Eritrea",
    "ES": "Spain",
    "ET": "Ethiopia",
    "FI": "Finland",
    "FJ": "Fiji",
    "FK": "Falkland Islands (Malvinas)",
    "FM": "Micronesia, Federated States of Micronesia",
    "FO": "Faroe Islands",
    "FR": "France",
    "FX": "France, Metropolitan",
    "GA": "Gabon",
    "GB": "United Kingdom",
    "GD": "Grenada",
    "GE": "Georgia",
    "GF": "French Guiana",
    "GH": "Ghana",
    "GI": "Gibraltar",
    "GL": "Greenland",
    "GM": "Gambia",
    "GN": "Guinea",
    "GP": "Guadeloupe",
    "GQ": "Equatorial Guinea",
    "GR": "Greece",
    "GS": "South Georgia and the South Sandwich Islands",
    "GT": "Guatemala",
    "GU": "Guam",
    "GW": "Guinea-Bissau",
    "GY": "Guyana",
    "HK": "Hong Kong",
    "HM": "Heard Island and McDonald Islands",
    "HN": "Honduras",
    "HR": "Croatia",
    "HT": "Haiti",
    "HU": "Hungary",
    "ID": "Indonesia",
    "IE": "Ireland",
    "IL": "Israel",
    "IN": "India",
    "IO": "British Indian Ocean Territory",
    "IQ": "Iraq",
    "IR": "Iran, Islamic Republic of Iran",
    "IS": "Iceland",
    "IT": "Italy",
    "JM": "Jamaica",
    "JO": "Jordan",
    "JP": "Japan",
    "KE": "Kenya",
    "KG": "Kyrgyzstan",
    "KH": "Cambodia",
    "KI": "Kiribati",
    "KM": "Comoros",
    "KN": "Saint Kitts and Nevis",
    "KP": "Korea, Democratic People's Republic of Korea",
    "KR": "Korea, Republic of Korea",
    "KW": "Kuwait",
    "KY": "Cayman Islands",
    "KZ": "Kazakhstan",
    "LA": "Lao People's Democratic Republic",
    "LB": "Lebanon",
    "LC": "Saint Lucia",
    "LI": "Liechtenstein",
    "LK": "Sri Lanka",
    "LR": "Liberia",
    "LS": "Lesotho",
    "LT": "Lithuania",
    "LU": "Luxembourg",
    "LV": "Latvia",
    "LY": "Libyan Arab Jamahiriya",
    "MA": "Morocco",
    "MC": "Monaco",
    "MD": "Moldova, Republic of Moldova",
    "MG": "Madagascar",
    "MH": "Marshall Islands",
    "MK": "Macedonia, the Former Yugoslav Republic of Macedonia",
    "ML": "Mali",
    "MM": "Myanmar",
    "MN": "Mongolia",
    "MO": "Macau",
    "MP": "Northern Mariana Islands",
    "MQ": "Martinique",
    "MR": "Mauritania",
    "MS": "Montserrat",
    "MT": "Malta",
    "MU": "Mauritius",
    "MV": "Maldives",
    "MW": "Malawi",
    "MX": "Mexico",
    "MY": "Malaysia",
    "MZ": "Mozambique",
    "NA": "Namibia",
    "NC": "New Caledonia",
    "NE": "Niger",
    "NF": "Norfolk Island",
    "NG": "Nigeria",
    "NI": "Nicaragua",
    "NL": "Netherlands",
    "NO": "Norway",
    "NP": "Nepal",
    "NR": "Nauru",
    "NU": "Niue",
    "NZ": "New Zealand",
    "OM": "Oman",
    "PA": "Panama",
    "PE": "Peru",
    "PF": "French Polynesia",
    "PG": "Papua New Guinea",
    "PH": "Philippines",
    "PK": "Pakistan",
    "PL": "Poland",
    "PM": "Saint Pierre and Miquelon",
    "PN": "Pitcairn",
    "PR": "Puerto Rico",
    "PS": "Palestinian Territory, Occupied",
    "PT": "Portugal",
    "PW": "Palau",
    "PY": "Paraguay",
    "QA": "Qatar",
    "RE": "Réunion",
    "RO": "Romania",
    "RU": "Russian Federation",
    "RW": "Rwanda",
    "SA": "Saudi Arabia",
    "SB": "Solomon Islands",
    "SC": "Seychelles",
    "SD": "Sudan",
    "SE": "Sweden",
    "SG": "Singapore",
    "SH": "Saint Helena, Ascension and Tristan da Cunha",
    "SI": "Slovenia",
    "SJ": "Svalbard and Jan Mayen",
    "SK": "Slovakia",
    "SL": "Sierra Leone",
    "SM": "San Marino",
    "SN": "Senegal",
    "SO": "Somalia",
    "SR": "Suriname",
    "ST": "Sao Tome and Principe",
    "SV": "El Salvador",
    "SY": "Syrian Arab Republic",
    "SZ": "Eswatini",
    "TC": "Turks and Caicos Islands",
    "TD": "Chad",
    "TF": "French Southern Territories",
    "TG": "Togo",
    "TH": "Thailand",
    "TJ": "Tajikistan",
    "TK": "Tokelau",
    "TM": "Turkmenistan",
    "TN": "Tunisia",
    "TO": "Tonga",
    "TL": "Timor-Leste",
    "TR": "Turkey",
    "TT": "Trinidad and Tobago",
    "TV": "Tuvalu",
    "TW": "Taiwan",
    "TZ": "Tanzania",
    "UA": "Ukraine",
    "UG": "Uganda",
    "UM": "United States Minor Outlying Islands",
    "US": "United States",
    "UY": "Uruguay",
    "UZ": "Uzbekistan",
    "VA": "Holy See",
    "VC": "Saint Vincent",
    "VE": "Venezuela",
    "VG": "Virgin Islands, British",
    "VI": "Virgin Islands, U.S.",
    "VN": "Vietnam",
    "VU": "Vanuatu",
    "WF": "Wallis and Futuna",
    "WS": "Samoa",
    "YE": "Yemen",
    "YT": "Mayotte",
    "RS": "Serbia",
    "ZA": "South Africa",
    "ZM": "Zambia",
    "ME": "Montenegro",
    "ZW": "Zimbabwe",
    "A1": "Unknown",
    "A2": "Satellite Provider",
    "O1": "Other",
    "AX": "Aland Islands",
    "GG": "Guernsey",
    "IM": "Isle of Man",
    "JE": "Jersey",
    "BL": "St. Barthelemy",
    "MF": "Saint Martin",
}

CountryNames = list(Countries.values())
CountryAcronyms = list(Countries.keys())




from typing import Iterable, Tuple
from .b1365 import b1365
from ..constants import *
from ..types import *
from ..io import *

class b1600(b1365):
    """
    b1600 adds /away commands.
    """
    version = 1600

    @classmethod
    def read_set_irc_away_message(cls, stream: MemoryStream) -> Message:
        return cls.read_message(stream)




from typing import Iterable, Tuple, Union
from .b20141104 import b20141104
from ..constants import *
from ..types import *
from ..io import *

class b20150826(b20141104):
    """
    b20150826 adds the "Qualified" status to the beatmap info reply.
    """
    version = 20150826

    @classmethod
    def convert_ranked_status(cls, status: Union[RankedStatus, int]) -> int:
        if type(status) is int:
            # A custom status was sent
            return status

        # Loved status does not exist
        status_mapping = {
            RankedStatus.NotSubmitted: -1,
            RankedStatus.Pending: 0,
            RankedStatus.Ranked: 1,
            RankedStatus.Approved: 2,
            RankedStatus.Qualified: 3,
            RankedStatus.Loved: 2
        }
        return status_mapping.get(status, -1)




from typing import Iterable, Tuple
from .b1817 import b1817
from ..constants import *
from ..types import *
from ..io import *

class b1820(b1817):
    """
    b1820 adds the image url to the title update packet.
    """
    version = 1820

    @classmethod
    def write_title_update(cls, update: TitleUpdate) -> Iterable[Tuple[PacketType, bytes]]:
        if not update.image_url:
            return []

        stream = MemoryStream()
        write_string(stream, update.image_url)
        yield PacketType.BanchoTitleUpdate, stream.data




from typing import Iterable, Tuple
from .b20121028 import b20121028
from ..constants import *
from ..types import *
from ..io import *

class b20121203(b20121028):
    """
    b20121203 optimizes the way how the user presence is sent to the client.
    """
    version = 20121203
    protocol_version = 13

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, cls.convert_user_id(info))
        write_string(stream, info.name)
        write_u8(stream, info.presence.timezone+24)
        write_u8(stream, info.presence.country_index)
        write_u8(stream, info.presence.permissions | info.status.mode << 5)
        write_f32(stream, info.presence.longitude)
        write_f32(stream, info.presence.latitude)
        write_s32(stream, info.stats.rank)
        write_u8(stream, info.status.mode)
        yield PacketType.BanchoUserPresence, stream.data




from typing import Iterable, Tuple
from .b20151106 import b20151106
from ..constants import *
from ..types import *
from ..io import *

class b20151107(b20151106):
    """
    b20151107 adds tournament client chat support.
    """
    version = 20151107

    @classmethod
    def read_tournament_join_match_channel(cls, stream: MemoryStream) -> int:
        return read_s32(stream)

    @classmethod
    def read_tournament_leave_match_channel(cls, stream: MemoryStream) -> int:
        return read_s32(stream)




from .clients import ClientDict, LowestVersion, HighestVersion
from .constants import CountryAcronyms
from .chio import BanchoIO

def resolve_country_index(country_acronym: str) -> int:
    """
    Resolve the country index from the acronym.
    If the acronym is not found, it will return 0.
    """
    return (
        CountryAcronyms.index(country_acronym)
        if country_acronym in CountryAcronyms else 0
    )

def select_client(version: int) -> BanchoIO:
    """Select the appropriate client based on the version provided."""
    if version in ClientDict:
        return ClientDict[version]
    
    if version < LowestVersion:
        return ClientDict[LowestVersion]
    
    if version > HighestVersion:
        return ClientDict[HighestVersion]
    
    for client_version, client in ClientDict.items():
        if version < client_version:
            return client

    # This should never happen, but just in case
    return ClientDict[HighestVersion]

def select_latest_client() -> BanchoIO:
    """Select the latest client available."""
    return ClientDict[HighestVersion]

def select_initial_client() -> BanchoIO:
    """Select the oldest client available."""
    return ClientDict[LowestVersion]




from typing import Iterable, Tuple
from .b452 import b452
from ..constants import *
from ..types import *
from ..io import *

class b470(b452):
    """
    b470 adds support for multiple game modes in multiplayer matches.
    """
    version = 470

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        stream = MemoryStream()
        write_u8(stream, match.id)
        write_boolean(stream, match.in_progress)
        write_u8(stream, match.type)
        write_u16(stream, match.mods.value)
        write_string(stream, match.name)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)

        for slot in match.slots:
            write_u8(stream, slot.status.value)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        write_s32(stream, match.host_id)
        write_u8(stream, match.mode)
        return stream.data

    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u8(stream)
        match.in_progress = read_boolean(stream)
        match.type = MatchType(read_u8(stream))
        match.mods = Mods(read_u16(stream))
        match.name = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)
        match.slots = [
            MatchSlot(status=SlotStatus(read_u8(stream)))
            for _ in range(cls.slot_size)
        ]

        for slot in match.slots:
            if slot.has_player:
                slot.user_id = read_s32(stream)

        match.host_id = read_s32(stream)
        match.mode = Mode(read_u8(stream))
        return match




from copy import copy
from typing import Iterable, Tuple
from .b354 import b354
from ..constants import *
from ..types import *
from ..io import *

class b365(b354):
    """
    b365 adds a level display on the user panel, which has a bug that causes
    the client to crash, when the user has a very high total score.
    """
    version = 365

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        total_score = copy(info.stats.tscore)

        # Cap total score & write stats
        info.stats.tscore = min(info.stats.tscore, 17705429348)
        yield next(super().write_user_stats(info))

        # Re-apply original value
        info.stats.tscore = total_score
    
    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        total_score = copy(info.stats.tscore)

        # Cap total score & write presence
        info.stats.tscore = min(info.stats.tscore, 17705429348)
        yield next(super().write_user_presence(info))

        # Re-apply original value
        info.stats.tscore = total_score




from typing import Iterable, Tuple
from .b291 import b291
from ..constants import *
from ..types import *
from ..io import *

class b294(b291):
    """
    b294 implements private messages, as well as score frames in spectating.
    """
    version = 294

    @classmethod
    def write_message(cls, message: Message) -> Iterable[Tuple[PacketType, bytes]]:
        if not message.is_direct_message and message.target != "#osu":
            # Channel selection has not been implemented yet
            return []

        stream = MemoryStream()
        write_string(stream, message.sender)
        write_string(stream, message.content)
        write_boolean(stream, message.is_direct_message)
        yield PacketType.BanchoMessage, stream.data

    @classmethod
    def read_message(cls, stream: MemoryStream):
        # Channel selection has not been implemented yet
        return Message(
            content=read_string(stream),
            target="#osu",
            sender=""
        )

    @classmethod
    def read_private_message(cls, stream: MemoryStream) -> Message:
        target = read_string(stream)
        content = read_string(stream)
        is_direct_message = read_boolean(stream)

        if not is_direct_message:
            raise ValueError("Expected direct message, got channel message")

        return Message("", content, target)

    @classmethod
    def write_spectate_frames(cls, bundle: ReplayFrameBundle) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u16(stream, len(bundle.frames))

        for frame in bundle.frames:
            stream.write(cls.write_replay_frame(frame))

        write_u8(stream, bundle.action)

        if bundle.frame:
            cls.write_score_frame(stream, bundle.frame)

        yield PacketType.BanchoSpectateFrames, stream.data

    @classmethod
    def read_spectate_frames(cls, stream: MemoryStream) -> ReplayFrameBundle:
        frames = [
            cls.read_replay_frame(stream)
            for _ in range(read_u16(stream))
        ]
        action = ReplayAction(read_u8(stream))
        frame = None

        if stream.available() > 0:
            frame = cls.read_score_frame(stream)

        return ReplayFrameBundle(action, frames, frame)

    @classmethod
    def write_score_frame(cls, stream: MemoryStream, frame: ScoreFrame) -> None:
        write_string(stream, frame.checksum)
        write_u8(stream, frame.id)
        write_u16(stream, frame.total_300)
        write_u16(stream, frame.total_100)
        write_u16(stream, frame.total_50)
        write_u16(stream, frame.total_geki)
        write_u16(stream, frame.total_katu)
        write_u16(stream, frame.total_miss)
        write_u32(stream, frame.total_score)
        write_u16(stream, frame.max_combo)
        write_u16(stream, frame.current_combo)
        write_boolean(stream, frame.perfect)
        write_u8(stream, frame.hp)

    @classmethod
    def read_score_frame(cls, stream: MemoryStream) -> ScoreFrame:
        # TODO: Validate checksum
        frame_checksum = read_string(stream)

        return ScoreFrame(
            time=0,
            id=read_u8(stream),
            total_300=read_u16(stream),
            total_100=read_u16(stream),
            total_50=read_u16(stream),
            total_geki=read_u16(stream),
            total_katu=read_u16(stream),
            total_miss=read_u16(stream),
            total_score=read_u32(stream),
            max_combo=read_u16(stream),
            current_combo=read_u16(stream),
            perfect=read_boolean(stream),
            hp=read_u8(stream),
            tag_byte=0
        )




from typing import Iterable, Tuple
from .b487 import b487
from ..constants import *
from ..types import *
from ..io import *

class b489(b487):
    """
    b489 adds support for game modes inside user status updates.
    """
    version = 489
    protocol_version = 1

    @classmethod
    def write_status_update(cls, status: UserStatus) -> bytes:
        stream = MemoryStream()
        write_u8(stream, status.action)

        beatmap_update = True
        write_boolean(stream, beatmap_update)

        if beatmap_update:
            write_string(stream, status.text)
            write_string(stream, status.beatmap_checksum)
            write_u16(stream, status.mods)

            if cls.protocol_version >= 1:
                write_u8(stream, status.mode)

        return stream.data

    @classmethod
    def read_user_status(cls, stream: MemoryStream) -> UserStatus:
        status = UserStatus()
        status.action = Status(read_u8(stream))
        beatmap_update = read_boolean(stream)

        if beatmap_update:
            status.text = read_string(stream)
            status.beatmap_checksum = read_string(stream)
            status.mods = Mods(read_u16(stream))

            if cls.protocol_version >= 1:
                status.mode = Mode(read_u8(stream))

        return status




from typing import Iterable, Tuple
from .b20140716 import b20140716
from ..constants import *
from ..types import *
from ..io import *

class b20140731(b20140716):
    """
    b20140731 adds the infamous "Zallius' eyes have awoken" (RTX) packet.
    """
    version = 20140731

    @classmethod
    def write_rtx(cls, message: str = "Zallius' eyes have awoken") -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, message)
        yield PacketType.BanchoRTX, stream.data




from typing import Iterable, Tuple, Any
from .b323 import b323
from ..constants import *
from ..types import *
from ..io import *

class b334(b323):
    """
    b334 introduces a lot of breaking changes:
    - Compression boolean inside packet header
    - Removal of checksums in score frames
    - Mods inside match struct
    - Packet IDs 50-58
    """
    version = 334
    header_size = 7

    @classmethod
    def read_packet(cls, stream: Stream) -> Tuple[PacketType, Any]:
        packet_id = read_u16(stream)
        packet = cls.convert_input_packet(packet_id)

        if not packet.is_client_packet:
            raise ValueError(f"Packet '{packet.name}' is not a client packet")

        packet_reader = getattr(cls, packet.handler_name, None)

        if not packet_reader:
            raise NotImplementedError(f"Version '{cls.version}' does not implement packet '{packet.name}'")

        compression = read_boolean(stream)
        packet_length = read_u32(stream)
        packet_data = stream.read(packet_length)

        if compression:
            packet_data = decompress(packet_data)

        return packet, packet_reader(MemoryStream(packet_data))

    @classmethod
    def write_packet(cls, stream: Stream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            compression_enabled = len(packet_data) > 150

            if compression_enabled and not cls.disable_compression:
                packet_data = compress(packet_data)

            write_u16(output_stream, packet_id)
            write_boolean(output_stream, compression_enabled)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            stream.write(output_stream.data)
            output_stream.clear()

    @classmethod
    async def read_packet_async(cls, stream: AsyncStream) -> Tuple[PacketType, Any]:
        input_stream = MemoryStream()
        input_stream.write(await stream.read(cls.header_size))

        packet_id = read_u16(input_stream)
        packet = cls.convert_input_packet(packet_id)

        if not packet.is_client_packet:
            raise ValueError(f"Packet '{packet.name}' is not a client packet")

        packet_reader = getattr(cls, packet.handler_name, None)

        if not packet_reader:
            raise NotImplementedError(f"Version '{cls.version}' does not implement packet '{packet.name}'")

        compression = read_boolean(input_stream)
        packet_length = read_u32(input_stream)
        packet_data = await stream.read(packet_length)

        if compression:
            packet_data = decompress(packet_data)

        return packet, packet_reader(MemoryStream(packet_data))

    @classmethod
    async def write_packet_async(cls, stream: AsyncStream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            compression_enabled = len(packet_data) > 150

            if compression_enabled:
                packet_data = compress(packet_data)

            write_u16(output_stream, packet_id)
            write_boolean(output_stream, compression_enabled)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            await stream.write(output_stream.data)
            output_stream.clear()

    @classmethod
    def convert_input_packet(cls, packet: int) -> PacketType:
        if packet == 11:
            # "IrcJoin" packet
            return PacketType.BanchoIrcJoin
        
        if packet == 51:
            # "MatchChangeBeatmap" packet
            return PacketType.OsuMatchChangeBeatmap

        if packet > 11:
            packet -= 1

        if packet >= 51:
            packet -= 1

        return PacketType(packet)

    @classmethod
    def convert_output_packet(cls, packet: PacketType) -> int:
        if packet == PacketType.BanchoIrcJoin:
            # "IrcJoin" packet
            return 11

        if packet == PacketType.OsuMatchChangeBeatmap:
            # "MatchChangeBeatmap" packet
            return 51

        packet_id = packet.value

        if packet_id >= 11:
            packet_id += 1

        if packet_id >= 51:
            packet_id += 1

        return packet_id

    @classmethod
    def write_match_transfer_host(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchTransferHost, b''

    @classmethod
    def write_match_all_players_loaded(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchAllPlayersLoaded, b''

    @classmethod
    def write_match_player_failed(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        yield PacketType.BanchoMatchPlayerFailed, stream.data

    @classmethod
    def write_match_complete(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchComplete, b''

    @classmethod
    def write_match_start(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchStart, cls.write_match(match)

    @classmethod
    def read_match_change_mods(cls, stream: MemoryStream) -> Mods:
        return Mods(read_s32(stream))

    @classmethod
    def read_match_load_complete(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_no_beatmap(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_not_ready(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_match_failed(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def write_score_frame(cls, stream: MemoryStream, frame: ScoreFrame) -> None:
        write_s32(stream, frame.time)
        write_u8(stream, frame.id)
        write_u16(stream, frame.total_300)
        write_u16(stream, frame.total_100)
        write_u16(stream, frame.total_50)
        write_u16(stream, frame.total_geki)
        write_u16(stream, frame.total_katu)
        write_u16(stream, frame.total_miss)
        write_u32(stream, frame.total_score)
        write_u16(stream, frame.max_combo)
        write_u16(stream, frame.current_combo)
        write_boolean(stream, frame.perfect)
        write_u8(stream, frame.hp)

    @classmethod
    def read_score_frame(cls, stream: MemoryStream) -> ScoreFrame:
        return ScoreFrame(
            time=read_s32(stream),
            id=read_u8(stream),
            total_300=read_u16(stream),
            total_100=read_u16(stream),
            total_50=read_u16(stream),
            total_geki=read_u16(stream),
            total_katu=read_u16(stream),
            total_miss=read_u16(stream),
            total_score=read_u32(stream),
            max_combo=read_u16(stream),
            current_combo=read_u16(stream),
            perfect=read_boolean(stream),
            hp=read_u8(stream),
            tag_byte=0
        )

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        stream = MemoryStream()
        write_u8(stream, match.id)
        write_boolean(stream, match.in_progress)
        write_u8(stream, match.type)
        write_u16(stream, match.mods.value)
        write_string(stream, match.name)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)

        for slot in match.slots:
            write_u8(stream, slot.status.value)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        return stream.data

    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u8(stream)
        match.in_progress = read_boolean(stream)
        match.type = MatchType(read_u8(stream))
        match.mods = Mods(read_u16(stream))
        match.name = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)
        match.slots = [
            MatchSlot(status=SlotStatus(read_u8(stream)))
            for _ in range(cls.slot_size)
        ]

        for slot in match.slots:
            if slot.has_player:
                slot.user_id = read_s32(stream)

        return match



from typing import Iterable, Tuple
from .b1183 import b1183
from ..constants import *
from ..types import *
from ..io import *

class b1365(b1183):
    """
    b1365 now sends "player skipped" messages in multiplayer.
    """
    version = 1365

    @classmethod
    def write_match_player_skipped(cls, slot_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, slot_id)
        yield PacketType.BanchoMatchPlayerSkipped, stream.data




from typing import Iterable, Tuple
from .b659 import b659
from ..constants import *
from ..types import *
from ..io import *

class b695(b659):
    """
    b695 changes the rank datatype from u16 to s32.
    """
    version = 695

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_u8(stream, Completeness.Statistics)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_u8(stream, Completeness.Full)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        write_string(stream, info.name)
        write_string(stream, info.avatar_filename)
        write_u8(stream, info.presence.timezone+24)
        write_string(stream, info.presence.country_string)
        write_u8(stream, info.presence.permissions)
        yield PacketType.BanchoUserStats, stream.data




from typing import Iterable, Tuple
from .b20130418 import b20130418
from ..constants import *
from ..types import *
from ..io import *

class b20130509(b20130418):
    """
    b20130509 reworks the updating system, and can now differentiate
    between forced updates and regular updates.
    """
    version = 20130509

    @classmethod
    def write_version_update_forced(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoVersionUpdateForced, b""




from typing import Iterable, Tuple
from .b1820 import b1820
from ..constants import *
from ..types import *
from ..io import *

class b20120518(b1820):
    """
    b20120518 introduced the "ChannelInfoComplete" packet.
    """
    version = 20120518

    @classmethod
    def write_channel_info_complete(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoChannelInfoComplete, b""




from typing import Optional, Iterable, Tuple
from .b470 import b470
from ..constants import *
from ..types import *
from ..io import *

class b487(b470):
    """
    b487 adds support for bancho protocol negotiations.
    """
    version = 487

    @classmethod
    def write_protocol_negotiation(cls, version: Optional[int] = None) -> Iterable[Tuple[PacketType, bytes]]:
        # This lets us decide if we want to use the pre-defined
        # version or a custom version we can provide.
        stream = MemoryStream()
        write_s32(stream, version or cls.protocol_version)
        yield PacketType.BanchoProtocolNegotiation, stream.data




from typing import Iterable, Tuple
from .b20121023 import b20121023
from ..constants import *
from ..types import *
from ..io import *

class b20121028(b20121023):
    """
    b20121028 adds the osu! mania ranking field to the beatmap info packet.
    """
    version = 20121028
    protocol_version = 12

    @classmethod
    def write_beatmap_info_reply(cls, reply: BeatmapInfoReply) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, len(reply.beatmaps))

        for info in reply.beatmaps:
            write_s16(stream, info.index)
            write_s32(stream, info.beatmap_id)
            write_s32(stream, info.beatmapset_id)
            write_s32(stream, info.thread_id)
            write_s8(stream, info.ranked_status)
            write_s8(stream, info.osu_rank)
            write_s8(stream, info.fruits_rank)
            write_s8(stream, info.taiko_rank)

            if cls.protocol_version >= 12:
                write_s8(stream, info.mania_rank)

            write_string(stream, info.checksum)

        yield PacketType.BanchoBeatmapInfoReply, stream.data




from typing import Callable
from functools import wraps
from .clients import ClientDict
from .constants import *

def patch(packet: PacketType, version: int) -> Callable:
    """
    Overwrite a packet handler for a specific protocol version.  
    For example, to overwrite the `BanchoUserStats` packet for b20120723:
    ```python
    @chio.patch(PacketType.BanchoUserStats, 20120723)
    def write_user_stats(cls, info: UserInfo):
        stream = MemoryStream()
        write_s32(stream, info.id)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        write_u16(stream, info.stats.pp)
        yield PacketType.BanchoUserStats, stream.data
    ```
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        client = ClientDict[version]
        setattr(client, packet.handler_name, wrapper)
        return wrapper
    return decorator

def set_protocol_version(protocol_version: int, version: int) -> None:
    """Override the protocol version for a specific client version."""
    client = ClientDict[version]
    client.protocol_version = protocol_version

def set_slot_size(slot_size: int, version: int) -> None:
    """Override the slot size for a specific client version."""
    client = ClientDict[version]
    client.slot_size = slot_size




from typing import Iterable, Tuple
from .b1600 import b1600
from ..constants import *
from ..types import *
from ..io import *

class b1788(b1600):
    """
    b1788 includes probably the largest amount of changes to the bancho protocol overall.
    - bUserStats split up into Presence & Stats
    - Server no longer needs to send bUserStats to clients, instead the client requests it themselves
    - The `count` datatype for bListInt got changed into a signed 16-bit integer
    - The user status no longer contains a boolean for updates
    - IrcJoin packet is now deprecated, in favor of user IDs being negative if they are IRC users
    - The restart packet got introduced, which informs the client that bancho is restarting
    - The exit packet now contains a boolean, which indicates if the client is exiting because of an update
    """
    version = 1788
    protocol_version = 6

    @classmethod
    def convert_input_packet(cls, packet: int) -> PacketType:
        return PacketType(packet)

    @classmethod
    def convert_output_packet(cls, packet: PacketType) -> int:
        return packet.value

    @classmethod
    def convert_user_id(cls, info: UserInfo) -> int:
        return -abs(info.id) if info.presence.is_irc else abs(info.id)

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, info.id)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, cls.convert_user_id(info))
        write_string(stream, info.name)
        write_u8(stream, AvatarExtension.Png)
        write_u8(stream, info.presence.timezone+24)
        write_string(stream, info.presence.country_string)
        write_u8(stream, info.presence.permissions)
        write_f32(stream, info.presence.longitude)
        write_f32(stream, info.presence.latitude)
        yield PacketType.BanchoUserPresence, stream.data

    @classmethod
    def read_user_status(cls, stream: MemoryStream) -> UserStatus:
        status = UserStatus()
        status.action = Status(read_u8(stream))
        status.text = read_string(stream)
        status.beatmap_checksum = read_string(stream)
        status.mods = Mods(read_u16(stream))
        status.mode = Mode(read_u8(stream))
        status.beatmap_id = read_s32(stream)
        return status

    @classmethod
    def write_status_update(cls, status: UserStatus) -> bytes:
        stream = MemoryStream()
        write_u8(stream, status.action)
        write_string(stream, status.text)
        write_string(stream, status.beatmap_checksum)
        write_u16(stream, status.mods)
        write_u8(stream, status.mode)
        write_s32(stream, status.beatmap_id)
        return stream.data

    @classmethod
    def write_friends_list(cls, friends: Iterable[int]) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_list_s16(stream, list(friends))
        yield PacketType.BanchoFriendsList, stream.data

    @classmethod
    def read_user_stats_request(cls, stream: MemoryStream) -> Iterable[int]:
        return read_list_s16(stream)

    @classmethod
    def write_restart(cls, retry_after_ms: int = 5000) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, retry_after_ms)
        yield PacketType.BanchoRestart, stream.data

    @classmethod
    def write_irc_join(cls, name: str) -> Iterable[Tuple[PacketType, bytes]]:
        return []
    
    @classmethod
    def write_irc_quit(cls, name: str) -> Iterable[Tuple[PacketType, bytes]]:
        return []
    
    @classmethod
    def read_exit(cls, stream: MemoryStream) -> bool:
        return read_s32(stream) == 1




from typing import Iterable, Tuple
from .b365 import b365
from ..constants import *
from ..types import *
from ..io import *

class b374(b365):
    """
    b374 adds the ButtonState constant, which deprectaes
    the old button left/right booleans.
    """
    version = 374

    @classmethod
    def write_replay_frame(cls, frame: ReplayFrame) -> bytes:
        stream = MemoryStream()
        write_u8(stream, frame.button_state)
        write_boolean(stream, False)
        write_f32(stream, frame.mouse_x)
        write_f32(stream, frame.mouse_y)
        write_s32(stream, frame.time)
        return stream.data

    @classmethod
    def read_replay_frame(cls, stream: MemoryStream) -> ReplayFrame:
        frame = ReplayFrame()
        frame.button_state = ButtonState(read_u8(stream))
        legacy_mouse_right = read_boolean(stream)
        frame.mouse_x = read_f32(stream)
        frame.mouse_y = read_f32(stream)
        frame.time = read_s32(stream)

        if legacy_mouse_right:
            frame.button_state |= ButtonState.Right1

        return frame




from typing import Iterable, Tuple, Any
from .b334 import b334
from ..constants import *
from ..types import *
from ..io import *

class b338(b334):
    """
    b338 changes the structure of statuses & stats.
    """
    version = 338

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_u8(stream, Completeness.Statistics)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u16(stream, info.stats.rank)
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_u8(stream, Completeness.Full)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u16(stream, info.stats.rank)
        write_string(stream, info.name)
        write_string(stream, info.avatar_filename)
        write_u8(stream, info.presence.timezone+24)
        write_string(stream, info.presence.country_string)
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_status_update(cls, status: UserStatus) -> bytes:
        stream = MemoryStream()
        write_u8(stream, status.action)

        beatmap_update = True
        write_boolean(stream, beatmap_update)

        if beatmap_update:
            write_string(stream, status.text)
            write_string(stream, status.beatmap_checksum)
            write_u16(stream, status.mods)

        return stream.data

    @classmethod
    def read_user_status(cls, stream: MemoryStream) -> UserStatus:
        status = UserStatus()
        status.action = Status(read_u8(stream))
        beatmap_update = read_boolean(stream)

        if beatmap_update:
            status.text = read_string(stream)
            status.beatmap_checksum = read_string(stream)
            status.mods = Mods(read_u16(stream))

        if status.action == Status.Idle and status.beatmap_checksum:
            # There is a bug where the client starts playing but
            # doesn't set the status to "Playing".
            status.action = Status.Playing

        return status




from typing import Iterable, Tuple
from .b20121224 import b20121224
from ..constants import *
from ..types import *
from ..io import *

class b20121225(b20121224):
    """
    b20121225 adds the user ID of the message sender to the message packet, so
    that the client can request the user's presence, if it's not available.
    """
    version = 20121225
    protocol_version = 15

    @classmethod
    def write_message(cls, message: Message) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, message.sender)
        write_string(stream, message.content)
        write_string(stream, message.target)

        if cls.protocol_version >= 15:
            write_s32(stream, message.sender_id)

        yield PacketType.BanchoMessage, stream.data

    @classmethod
    def read_message(cls, stream: MemoryStream) -> Message:
        message = Message(
            sender=read_string(stream),
            content=read_string(stream),
            target=read_string(stream)
        )

        if cls.protocol_version >= 15:
            message.sender_id = read_s32(stream)

        return message




from typing import Iterable, Tuple
from .b490 import b490
from ..constants import *
from ..types import *
from ..io import *

class b504(b490):
    """
    b504 adds the taiko & fruits ranks to the beatmap info packet.
    """
    version = 504
    protocol_version = 2

    @classmethod
    def write_beatmap_info_reply(cls, reply: BeatmapInfoReply) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, len(reply.beatmaps))

        for info in reply.beatmaps:
            write_s16(stream, info.index)
            write_s32(stream, info.beatmap_id)
            write_s32(stream, info.beatmapset_id)
            write_s32(stream, info.thread_id)
            write_s8(stream, cls.convert_ranked_status(info.ranked_status))
            write_s8(stream, info.osu_rank)

            if cls.protocol_version >= 2:
                write_s8(stream, info.fruits_rank)
                write_s8(stream, info.taiko_rank)

            write_string(stream, info.checksum)

        yield PacketType.BanchoBeatmapInfoReply, stream.data

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        # NOTE: See b365 for the level overflow bug
        write_u32(stream, info.id)
        write_u8(stream, Completeness.Statistics)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, min(info.stats.tscore, 26931190827))
        write_u16(stream, info.stats.rank)
        yield PacketType.BanchoUserStats, stream.data
    
    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        # NOTE: See b365 for the level overflow bug
        write_u32(stream, info.id)
        write_u8(stream, Completeness.Full)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, min(info.stats.tscore, 26931190827))
        write_u16(stream, info.stats.rank)
        write_string(stream, info.name)
        write_string(stream, info.avatar_filename)
        write_u8(stream, info.presence.timezone+24)
        write_string(stream, info.presence.country_string)
        write_u8(stream, info.presence.permissions)
        yield PacketType.BanchoUserStats, stream.data




from typing import Iterable, Tuple
from .b20130509 import b20130509
from ..constants import *
from ..types import *
from ..io import *

class b20130604(b20130509):
    """
    b20130604 allows for forced bancho server switching, after
    the client was idle for a certain amount of time.
    """
    version = 20130604

    @classmethod
    def write_switch_server(cls, after_idle_time: int = 3600) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(after_idle_time)
        yield PacketType.BanchoSwitchServer, stream.data




from typing import Iterable, Tuple
from ..types import PacketType
from .b282 import b282
from ..io import *

class b291(b282):
    """
    b291 implements the GetAttension & Announce packets.
    """
    version = 291

    @classmethod
    def write_get_attention(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoGetAttention, b''

    @classmethod
    def write_announce(cls, message: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, message)
        yield PacketType.BanchoAnnounce, stream.data

    @classmethod
    def write_restart(cls, retry_after_ms: int = 5000) -> Iterable[Tuple[PacketType, bytes]]:
        # NOTE: This is a backport of the actual restart packet, that
        #       simply announces the server restart to the user.
        return cls.write_announce(f"Bancho is restarting, please wait...")




from typing import List, Any, Tuple, Iterable, Union
from gzip import compress

from ..chio import BanchoIO
from ..constants import *
from ..types import *
from ..io import *

class b282(BanchoIO):
    """
    b282 is the initial implementation of the bancho protocol.
    Every following version will be based on it.
    """
    version = 282

    @classmethod
    def read_packet(cls, stream: Stream) -> Tuple[PacketType, Any]:
        packet_id = read_u16(stream)
        packet = cls.convert_input_packet(packet_id)

        if not packet.is_client_packet:
            raise ValueError(f"Packet '{packet.name}' is not a client packet")

        packet_reader = getattr(cls, packet.handler_name, None)

        if not packet_reader:
            raise NotImplementedError(f"Version '{cls.version}' does not implement packet '{packet.name}'")

        packet_length = read_u32(stream)
        packet_data = read_gzip(stream, packet_length)
        return packet, packet_reader(MemoryStream(packet_data))

    @classmethod
    def write_packet(cls, stream: Stream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            packet_data = compress(packet_data)
            write_u16(output_stream, packet_id)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            stream.write(output_stream.data)
            output_stream.clear()

    @classmethod
    async def read_packet_async(cls, stream: AsyncStream) -> Tuple[PacketType, Any]:
        input_stream = MemoryStream()
        input_stream.write(await stream.read(cls.header_size))

        packet_id = read_u16(input_stream)
        packet = cls.convert_input_packet(packet_id)

        if not packet.is_client_packet:
            raise ValueError(f"Packet '{packet.name}' is not a client packet")
        
        packet_reader = getattr(cls, packet.handler_name, None)

        if not packet_reader:
            raise NotImplementedError(f"Version '{cls.version}' does not implement packet '{packet.name}'")
        
        packet_length = read_u32(input_stream)
        packet_data = await stream.read(packet_length)
        packet_data = decompress(packet_data)
        return packet, packet_reader(MemoryStream(packet_data))

    @classmethod
    async def write_packet_async(cls, stream: AsyncStream, packet: PacketType, *args) -> None:
        if not packet.is_server_packet:
            raise ValueError(f"Packet '{packet.name}' is not a server packet")

        packet_writer = getattr(cls, packet.handler_name, None)

        if not packet_writer:
            return

        packets = packet_writer(*args)
        output_stream = MemoryStream()

        for packet, packet_data in packets:
            packet_id = cls.convert_output_packet(packet)
            packet_data = compress(packet_data)
            write_u16(output_stream, packet_id)
            write_u32(output_stream, len(packet_data))
            output_stream.write(packet_data)
            await stream.write(output_stream.data)
            output_stream.clear()

    @classmethod
    def convert_input_packet(cls, packet: int) -> PacketType:
        """
        Convert a packet from the client to a modern packet
        type that can be handled by the server.
        """
        if packet == 11:
            # "IrcJoin" packet
            return PacketType.BanchoIrcJoin

        if packet > 11 and packet <= 45:
            packet -= 1

        if packet > 50:
            packet -= 1

        return PacketType(packet)

    @classmethod
    def convert_output_packet(cls, packet: PacketType) -> int:
        """
        Convert a modern packet type from the server
        that the client can understand.
        """
        if packet is PacketType.BanchoIrcJoin:
            # "IrcJoin" packet
            return 11

        if packet >= 11 and packet < 45:
            return packet.value + 1

        if packet > 50:
            return packet.value + 1

        return packet.value

    @classmethod
    def convert_input_status(cls, status: int) -> Status:
        if status == 10:
            return Status.Unknown

        if status > 9:
            return Status(status - 1)

        return Status(status)

    @classmethod
    def convert_output_status(cls, status: UserStatus) -> Status:
        if status.update_stats:
            return Status.StatsUpdate

        if status.action > 9:
            return Status(status.action.value - 1)

        return status.action

    @classmethod
    def write_login_reply(cls, reply: Union[int, LoginError]) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, reply)
        yield PacketType.BanchoLoginReply, stream.data

    @classmethod
    def write_ping(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoPing, b""

    @classmethod
    def write_message(cls, message: Message) -> Iterable[Tuple[PacketType, bytes]]:
        if message.target != "#osu":
            # Private messages & channels have not been implemented yet
            return []

        stream = MemoryStream()
        write_string(stream, message.sender)
        write_string(stream, message.content)
        yield PacketType.BanchoMessage, stream.data

    @classmethod
    def write_irc_change_username(cls, old_name: str, new_name: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, f"{old_name}>>>>{new_name}")
        yield PacketType.BanchoIrcChangeUsername, stream.data

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()

        if info.presence.is_irc:
            yield next(cls.write_irc_join(info.name))
            return

        write_u32(stream, info.id)
        write_string(stream, info.name)
        write_u64(stream, info.stats.rscore)
        write_f64(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)
        write_string(stream, info.avatar_filename)
        stream.write(cls.write_status_update(info.status))
        write_u8(stream, info.presence.timezone+24)
        write_string(stream, info.presence.country_string)
        yield PacketType.BanchoUserStats, stream.data

    @classmethod
    def write_status_update(cls, status: UserStatus) -> bytes:
        action = cls.convert_output_status(status)
        stream = MemoryStream()
        write_u8(stream, action)

        if action != Status.Unknown:
            write_string(stream, status.text)
            write_string(stream, status.beatmap_checksum)
            write_u16(stream, status.mods)

        return stream.data

    @classmethod
    def write_user_quit(cls, quit: UserQuit) -> Iterable[Tuple[PacketType, bytes]]:
        if quit.info.presence.is_irc and quit.state != QuitState.IrcRemaining:
            stream = MemoryStream()
            write_string(stream, quit.info.name)
            return [(PacketType.BanchoIrcQuit, stream.data)]

        if quit.state == QuitState.OsuRemaining:
            return []

        packet, data = next(cls.write_user_stats(quit.info))
        packet = PacketType.BanchoUserQuit
        yield packet, data

    @classmethod
    def write_irc_join(cls, name: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, name)
        yield PacketType.BanchoIrcJoin, stream.data

    @classmethod
    def write_irc_quit(cls, name: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, name)
        yield PacketType.BanchoIrcQuit, stream.data

    @classmethod
    def write_spectator_joined(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, user_id)
        yield PacketType.BanchoSpectatorJoined, stream.data

    @classmethod
    def write_spectator_left(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, user_id)
        yield PacketType.BanchoSpectatorLeft, stream.data

    @classmethod
    def write_spectate_frames(cls, bundle: ReplayFrameBundle) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u16(stream, len(bundle.frames))

        for frame in bundle.frames:
            stream.write(cls.write_replay_frame(frame))

        write_u8(stream, bundle.action)
        yield PacketType.BanchoSpectateFrames, stream.data

    @classmethod
    def write_replay_frame(cls, frame: ReplayFrame) -> bytes:
        stream = MemoryStream()
        left_mouse = ButtonState.Left1 in frame.button_state or ButtonState.Left2 in frame.button_state
        right_mouse = ButtonState.Right1 in frame.button_state or ButtonState.Right2 in frame.button_state
        write_boolean(stream, left_mouse)
        write_boolean(stream, right_mouse)
        write_f32(stream, frame.mouse_x)
        write_f32(stream, frame.mouse_y)
        write_s32(stream, frame.time)
        return stream.data

    @classmethod
    def write_version_update(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoVersionUpdate, b""

    @classmethod
    def write_spectator_cant_spectate(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, user_id)
        yield PacketType.BanchoSpectatorCantSpectate, stream.data

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        # b282 does not support user presences,
        # instead we will send a stats update
        return cls.write_user_stats(info)

    @classmethod
    def write_user_presence_single(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        return cls.write_user_presence(info)

    @classmethod
    def write_user_presence_bundle(cls, infos: List[UserInfo]) -> Iterable[Tuple[PacketType, bytes]]:
        for info in infos:
            for packet in cls.write_user_presence_single(info):
                yield packet

    @classmethod
    def write_invite(cls, invite_message: Message) -> Iterable[Tuple[PacketType, bytes]]:
        # b282 does not support invites, so instead
        # we will send the message directly
        return cls.write_message(invite_message)

    @classmethod
    def read_user_status(cls, stream: MemoryStream) -> UserStatus:
        status = UserStatus()
        status.action = cls.convert_input_status(read_u8(stream))

        if status.action != Status.Unknown:
            status.text = read_string(stream)
            status.beatmap_checksum = read_string(stream)
            status.mods = Mods(read_u16(stream))

        if status.action == Status.Idle and status.beatmap_checksum:
            # There is a bug where the client starts playing but
            # doesn't set the status to "Playing".
            status.action = Status.Playing

        return status

    @classmethod
    def read_message(cls, stream: MemoryStream) -> Message:
        # Private messages & channels have not been implemented yet
        return Message(
            content=read_string(stream),
            target="#osu",
            sender=""
        )

    @classmethod
    def read_exit(cls, stream: MemoryStream) -> bool:
        return False

    @classmethod
    def read_status_update_request(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_pong(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_start_spectating(cls, stream: MemoryStream) -> int:
        return read_s32(stream)

    @classmethod
    def read_stop_spectating(cls, stream: MemoryStream) -> None:
        pass

    @classmethod
    def read_spectate_frames(cls, stream: MemoryStream) -> ReplayFrameBundle:
        frames = [
            cls.read_replay_frame(stream)
            for _ in range(read_u16(stream))
        ]
        action = ReplayAction(read_u8(stream))
        return ReplayFrameBundle(action, frames)

    @classmethod
    def read_replay_frame(cls, stream: MemoryStream) -> ReplayFrame:
        frame = ReplayFrame()
        mouse_left = read_boolean(stream)
        mouse_right = read_boolean(stream)
        frame.mouse_x = read_f32(stream)
        frame.mouse_y = read_f32(stream)
        frame.time = read_s32(stream)

        if mouse_left:
            frame.button_state |= ButtonState.Left1
        if mouse_right:
            frame.button_state |= ButtonState.Right1

        return frame

    @classmethod
    def read_error_report(cls, stream: MemoryStream) -> str:
        return read_string(stream)

    @classmethod
    def read_cant_spectate(cls, stream: MemoryStream) -> None:
        pass




from typing import Iterable, Tuple
from .b489 import b489
from ..constants import *
from ..types import *
from ..io import *

class b490(b489):
    """
    b490 now sends the beatmap ID in user status updates and
    can request a list of set IDs inside of the beatmap info request.
    """
    version = 490

    @classmethod
    def write_status_update(cls, status: UserStatus) -> bytes:
        stream = MemoryStream()
        write_u8(stream, status.action)

        beatmap_update = True
        write_boolean(stream, beatmap_update)

        if beatmap_update:
            write_string(stream, status.text)
            write_string(stream, status.beatmap_checksum)
            write_u16(stream, status.mods)
            write_u8(stream, status.mode)
            write_s32(stream, status.beatmap_id)

        return stream.data

    @classmethod
    def read_user_status(cls, stream: MemoryStream) -> UserStatus:
        status = UserStatus()
        status.action = Status(read_u8(stream))
        beatmap_update = read_boolean(stream)

        if beatmap_update:
            status.text = read_string(stream)
            status.beatmap_checksum = read_string(stream)
            status.mods = Mods(read_u16(stream))
            status.mode = Mode(read_u8(stream))
            status.beatmap_id = read_s32(stream)

        return status
    
    @classmethod
    def read_beatmap_info_request(cls, stream: MemoryStream) -> BeatmapInfoRequest:
        return BeatmapInfoRequest(
            [read_string(stream) for _ in range(read_u32(stream))],
            [read_s32(stream) for _ in range(read_u32(stream))]
        )




from typing import Iterable, Tuple, Union
from .b20130604 import b20130604
from ..constants import *
from ..types import *
from ..io import *

class b20130801(b20130604):
    """
    b20130801 deprecates the "user is inactive" error, and is
    now interpreted as a "user is banned" error by clients.
    Chio will try to account for this by sending an announce
    packet with the original message.
    """
    version = 20130801

    @classmethod
    def write_login_reply(cls, reply: Union[int, LoginError]) -> Iterable[Tuple[PacketType, bytes]]:
        if reply is LoginError.UserInactive:
            # Clients interpret this as a "user is banned" error,
            # instead of the intended "user is inactive" error.
            reply = LoginError.InvalidLogin
            yield next(cls.write_announce(InactiveAccountMessage))

        yield next(super().write_login_reply(reply))




from typing import Iterable, Tuple
from .b20140528 import b20140528
from ..constants import *
from ..types import *
from ..io import *

class b20140716(b20140528):
    """
    b20140716 sends the restriction status to the client.
    """
    version = 20140716

    @classmethod
    def write_account_restricted(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoAccountRestricted, b""




from typing import Iterable, Tuple
from .b20120518 import b20120518
from ..constants import *
from ..types import *
from ..io import *

class b20120703(b20120518):
    """
    b20120703 introduces a new packet to change match passwords.
    """
    version = 20120703

    @classmethod
    def read_match_change_password(cls, stream: MemoryStream) -> Match:
        return cls.read_match(stream)

    @classmethod
    def write_match_change_password(cls, password: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, password)
        yield PacketType.BanchoMatchChangePassword, stream.data




from .b294 import b294
from ..constants import *
from ..types import *
from ..io import *

class b296(b294):
    """
    b296 adds the "Time" value to score frames.
    """
    version = 296

    @classmethod
    def write_score_frame(cls, stream: MemoryStream, frame: ScoreFrame) -> None:
        write_string(stream, frame.checksum)
        write_s32(stream, frame.time)
        write_u8(stream, frame.id)
        write_u16(stream, frame.total_300)
        write_u16(stream, frame.total_100)
        write_u16(stream, frame.total_50)
        write_u16(stream, frame.total_geki)
        write_u16(stream, frame.total_katu)
        write_u16(stream, frame.total_miss)
        write_u32(stream, frame.total_score)
        write_u16(stream, frame.max_combo)
        write_u16(stream, frame.current_combo)
        write_boolean(stream, frame.perfect)
        write_u8(stream, frame.hp)

    @classmethod
    def read_score_frame(cls, stream):
        # TODO: Validate checksum
        frame_checksum = read_string(stream)

        return ScoreFrame(
            time=read_s32(stream),
            id=read_u8(stream),
            total_300=read_u16(stream),
            total_100=read_u16(stream),
            total_50=read_u16(stream),
            total_geki=read_u16(stream),
            total_katu=read_u16(stream),
            total_miss=read_u16(stream),
            total_score=read_u32(stream),
            max_combo=read_u16(stream),
            current_combo=read_u16(stream),
            perfect=read_boolean(stream),
            hp=read_u8(stream),
            tag_byte=0
        )




from typing import Iterable, Tuple
from .b20120723 import b20120723
from ..constants import *
from ..types import *
from ..io import *

class b20120725(b20120723):
    """
    b20120725 now displays the topic & user count of a channel in the channel listing.
    """
    version = 20120725
    protocol_version = 9

    @classmethod
    def write_channel_available(cls, channel: Channel) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel.name)

        if cls.protocol_version >= 9:
            write_string(stream, channel.topic)
            write_s16(stream, channel.user_count)

        yield PacketType.BanchoChannelAvailable, stream.data

    @classmethod
    def write_channel_available_autojoin(cls, channel: Channel) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel.name)

        if cls.protocol_version >= 9:
            write_string(stream, channel.topic)
            write_s16(stream, channel.user_count)

        yield PacketType.BanchoChannelAvailableAutojoin, stream.data




from typing import Iterable, Tuple
from .b20121211 import b20121211
from ..constants import *
from ..types import *
from ..io import *

class b20121212(b20121211):
    """
    b20121212 adds support for the user silenced packet.
    """
    version = 20121212

    @classmethod
    def write_user_silenced(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, user_id)
        yield PacketType.BanchoUserSilenced, stream.data




from typing import Iterable, Tuple
from .b1788 import b1788
from ..constants import *
from ..types import *
from ..io import *

class b1796(b1788):
    """
    b1796 changes the way how countries are sent to the client, and increases
    the total amount of match-slots that can be sent (u8 -> u16).
    """
    version = 1796

    @classmethod
    def write_match_update(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchUpdate, cls.write_match(match)

    @classmethod
    def write_match_new(cls, match: Match) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMatchNew, cls.write_match(match)

    @classmethod
    def write_user_presence(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, cls.convert_user_id(info))
        write_string(stream, info.name)
        write_u8(stream, AvatarExtension.Png)
        write_u8(stream, info.presence.timezone+24)
        write_u8(stream, info.presence.country_index)
        write_string(stream, info.presence.city)
        write_u8(stream, info.presence.permissions)
        write_f32(stream, info.presence.longitude)
        write_f32(stream, info.presence.latitude)
        yield PacketType.BanchoUserPresence, stream.data

    @classmethod
    def write_match(cls, match: Match) -> bytes:
        stream = MemoryStream()
        write_u16(stream, match.id)
        write_boolean(stream, match.in_progress)
        write_u8(stream, match.type)
        write_u16(stream, match.mods.value)
        write_string(stream, match.name)
        write_string(stream, match.password)
        write_string(stream, match.beatmap_text)
        write_s32(stream, match.beatmap_id)
        write_string(stream, match.beatmap_checksum)

        for slot in match.slots:
            write_u8(stream, slot.status.value)

        if cls.protocol_version >= 4:
            for slot in match.slots:
                write_u8(stream, slot.team)

        for slot in match.slots:
            if slot.has_player:
                write_s32(stream, slot.user_id)

        write_s32(stream, match.host_id)
        write_u8(stream, match.mode)

        if cls.protocol_version >= 3:
            write_u8(stream, match.scoring_type)
            write_u8(stream, match.team_type)

        return stream.data

    @classmethod
    def read_match(cls, stream: MemoryStream) -> Match:
        match = Match()
        match.id = read_u16(stream)
        match.in_progress = read_boolean(stream)
        match.type = MatchType(read_u8(stream))
        match.mods = Mods(read_u16(stream))
        match.name = read_string(stream)
        match.password = read_string(stream)
        match.beatmap_text = read_string(stream)
        match.beatmap_id = read_s32(stream)
        match.beatmap_checksum = read_string(stream)
        match.slots = [
            MatchSlot(status=SlotStatus(read_u8(stream)))
            for _ in range(cls.slot_size)
        ]

        if cls.protocol_version >= 4:
            for slot in match.slots:
                slot.team = SlotTeam(read_u8(stream))

        for slot in match.slots:
            if slot.has_player:
                slot.user_id = read_s32(stream)

        match.host_id = read_s32(stream)
        match.mode = Mode(read_u8(stream))

        if cls.protocol_version < 3:
            return match

        match.scoring_type = ScoringType(read_u8(stream))
        match.team_type = TeamType(read_u8(stream))
        return match




from typing import Iterable, Tuple
from .b20120703 import b20120703
from ..constants import *
from ..types import *
from ..io import *

class b20120723(b20120703):
    """
    b20120723 now shows your performance points in-game.
    """
    version = 20120723
    protocol_version = 8

    @classmethod
    def write_user_stats(cls, info: UserInfo) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, info.id)
        stream.write(cls.write_status_update(info.status))
        write_u64(stream, info.stats.rscore)
        write_f32(stream, info.stats.accuracy)
        write_u32(stream, info.stats.playcount)
        write_u64(stream, info.stats.tscore)
        write_u32(stream, info.stats.rank)

        if cls.protocol_version >= 8:
            write_s16(stream, info.stats.pp)

        yield PacketType.BanchoUserStats, stream.data




from typing import Iterable, Tuple
from .b312 import b312
from ..constants import *
from ..types import *
from ..io import *

class b320(b312):
    """
    b320 adds partial support for multiple channels
    """
    version = 320

    @classmethod
    def write_message(cls, message) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, message.sender)
        write_string(stream, message.content)
        write_string(stream, message.target)
        yield PacketType.BanchoMessage, stream.data

    @classmethod
    def read_message(cls, stream: MemoryStream) -> Message:
        return Message(
            sender=read_string(stream),
            content=read_string(stream),
            target=read_string(stream)
        )
    
    @classmethod
    def read_private_message(cls, stream: MemoryStream) -> Message:
        return cls.read_message(stream)




from typing import Iterable, Tuple
from .b20121221 import b20121221
from ..constants import *
from ..types import *
from ..io import *

class b20121224(b20121221):
    """
    b20121224 deprecates the irc quit packet, in favor of a unified quit packet.
    Additionally it changes the way how user presences are sent to the client,
    when the client first joins the server, to reduce bandwidth usage.
    """
    version = 20121224

    @classmethod
    def write_user_quit(cls, quit: UserQuit) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, quit.info.id)
        write_u8(stream, quit.state)
        yield PacketType.BanchoUserQuit, stream.data

    @classmethod
    def write_user_presence_bundle(cls, infos: List[UserInfo]) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_list_s16(stream, [info.id for info in infos])
        yield PacketType.BanchoUserPresenceBundle, stream.data

    @classmethod
    def write_user_presence_single(cls, info):
        stream = MemoryStream()
        write_s32(stream, info.id)
        yield PacketType.BanchoUserPresenceSingle, stream.data

    @classmethod
    def read_presence_request(cls, stream: MemoryStream) -> List[int]:
        return read_list_s16(stream)

    @classmethod
    def read_presence_request_all(cls, stream: MemoryStream) -> None:
        pass




from typing import Iterable, Tuple
from .b342 import b342
from ..constants import *
from ..types import *
from ..io import *

class b349(b342):
    """
    b349 adds the channel datatype and packets.
    """
    version = 349

    @classmethod
    def read_channel_join(cls, stream: MemoryStream) -> str:
        return read_string(stream)

    @classmethod
    def write_channel_join_success(cls, channel: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel)
        yield PacketType.BanchoChannelJoinSuccess, stream.data

    @classmethod
    def write_channel_revoked(cls, channel: str) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel)
        yield PacketType.BanchoChannelRevoked, stream.data

    @classmethod
    def write_channel_available(cls, channel: Channel) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel.name)
        yield PacketType.BanchoChannelAvailable, stream.data

    @classmethod
    def write_channel_available_autojoin(cls, channel: Channel) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_string(stream, channel.name)
        yield PacketType.BanchoChannelAvailableAutojoin, stream.data




from typing import Iterable, Tuple
from .b402 import b402
from ..constants import *
from ..types import *
from ..io import *

class b425(b402):
    """
    b425 adds support for user permissions.
    """
    version = 425

    @classmethod
    def write_login_permissions(cls, permissions: Permissions) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_s32(stream, permissions.value)
        yield PacketType.BanchoLoginPermissions, stream.data




import setuptools
import chio
import os

current_directory = os.path.dirname(os.path.abspath(__file__))
readme_path = os.path.join(current_directory, "README.md")

with open(readme_path, "r") as f:
    long_description = f.read()

setuptools.setup(
    name="chio",
    version=chio.__version__,
    author=chio.__author__,
    author_email=chio.__email__,
    description="A python library for serializing and deserializing bancho packets.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    keywords=["osu", "osugame", "python", "bancho"],
)




from typing import Iterable, Tuple, Union
from .b374 import b374
from ..constants import *
from ..types import *
from ..io import *

class b388(b374):
    """
    b388 changes ranked status from bool->int in beatmap info packets.
    """
    version = 388

    @classmethod
    def convert_ranked_status(cls, status: Union[RankedStatus, int]) -> int:
        if type(status) is int:
            # A custom status was sent
            return status

        # Approved status does not exist
        status_mapping = {
            RankedStatus.NotSubmitted: -1,
            RankedStatus.Pending: 0,
            RankedStatus.Ranked: 1,
            RankedStatus.Approved: 2,
            RankedStatus.Qualified: 2,
            RankedStatus.Loved: 2
        }
        return status_mapping.get(status, -1)

    @classmethod
    def write_beatmap_info_reply(cls, reply: BeatmapInfoReply) -> Iterable[Tuple[PacketType, bytes]]:
        stream = MemoryStream()
        write_u32(stream, len(reply.beatmaps))

        for info in reply.beatmaps:
            write_s16(stream, info.index)
            write_s32(stream, info.beatmap_id)
            write_s32(stream, info.beatmapset_id)
            write_s32(stream, info.thread_id)
            write_s8(stream, cls.convert_ranked_status(info.ranked_status))
            write_s8(stream, info.osu_rank)
            write_string(stream, info.checksum)

        yield PacketType.BanchoBeatmapInfoReply, stream.data




from typing import Iterable, Tuple
from .b20130801 import b20130801
from ..constants import *
from ..types import *
from ..io import *

class b20140528(b20130801):
    """
    b20140528 allows for 16-player multiplayer matches, if
    the protocol version is set to 19 or higher.
    """
    version = 20140528
    protocol_version = 18
    slot_size = 16

    @classmethod
    def read_match(cls, stream):
        cls.slot_size = 16 if cls.protocol_version >= 19 else 8
        return super().read_match(stream)

    @classmethod
    def write_match(cls, match):
        cls.slot_size = 16 if cls.protocol_version >= 19 else 8
        return super().write_match(match)




from gzip import decompress, compress
from abc import ABC, abstractmethod
from struct import pack, unpack
from typing import List

import traceback
import logging

class Stream(ABC):
    """
    Abstract class for I/O operations.
    """
    @abstractmethod
    def read(self, size: int = -1) -> bytes:
        """
        Read a number of bytes from the stream.
        """
        pass

    @abstractmethod
    def write(self, data: bytes) -> None:
        """
        Write a number of bytes to the stream.
        """
        pass

class AsyncStream(ABC):
    """
    Abstract class for asynchronous I/O operations.
    """
    @abstractmethod
    async def read(self, size: int = -1) -> bytes:
        """
        Read a number of bytes from the stream asynchronously.
        """
        pass

    @abstractmethod
    async def write(self, data: bytes) -> None:
        """
        Write a number of bytes to the stream asynchronously.
        """
        pass

class MemoryStream(Stream):
    """
    Stream implementation that uses an in-memory buffer.
    """

    def __init__(self, data: bytes = b"", endian: str = "<") -> None:
        self.data = data
        self.position = 0
        self.struct_endian = endian

    def __del__(self):
        self.clear()

    @property
    def endian(self) -> str:
        return self.struct_endian

    def write(self, data: bytes) -> None:
        self.data += data

    def read(self, size: int = -1) -> bytes:
        if size == -1:
            size = len(self.data) - self.position

        data = self.data[self.position:self.position + size]
        self.position += size
        return data
    
    def clear(self) -> None:
        self.data = b""
        self.position = 0

    def available(self) -> int:
        return len(self.data) - self.position

logger = logging.getLogger('chio.py')

def clamp(value: int, min_value: int, max_value: int) -> int:
    clamped = max(min_value, min(value, max_value))

    if clamped != value:
        stack = traceback.extract_stack()
        caller = stack[-3]
        logger.warning(
            f"Value '{value}' was clamped to '{clamped}' "
            f"({caller.filename}:{caller.lineno} in '{caller.name}')"
        )

    return clamped

def read_s8(stream: Stream) -> int:
    return stream.read(1)[0]

def read_u8(stream: Stream) -> int:
    return stream.read(1)[0]

def read_u16(stream: Stream) -> int:
    return unpack("<H", stream.read(2))[0]

def read_s16(stream: Stream) -> int:
    return unpack("<h", stream.read(2))[0]

def read_u32(stream: Stream) -> int:
    return unpack("<I", stream.read(4))[0]

def read_s32(stream: Stream) -> int:
    return unpack("<i", stream.read(4))[0]

def read_u64(stream: Stream) -> int:
    return unpack("<Q", stream.read(8))[0]

def read_s64(stream: Stream) -> int:
    return unpack("<q", stream.read(8))[0]

def read_boolean(stream: Stream) -> bool:
    return bool(read_u8(stream))

def read_f32(stream: Stream) -> float:
    return unpack("<f", stream.read(4))[0]

def read_f64(stream: Stream) -> float:
    return unpack("<d", stream.read(8))[0]

def read_gzip(stream: Stream, size: int = -1) -> bytes:
    return decompress(stream.read(size))

def read_uleb128(stream: Stream) -> int:
    num = shift = 0

    while True:
        byte = read_s8(stream)
        num |= (byte & 0x7F) << shift
        if (byte & 0x80) == 0:
            break

        shift += 7

    return num

def read_string(stream: Stream) -> str:
    empty = read_s8(stream) == 0x00

    if empty:
        return ""

    size = read_uleb128(stream)
    return stream.read(size).decode()

def read_bool_list(stream: Stream, size: int = 8) -> List[bool]:
    byte = read_u8(stream)
    return [((byte >> index) & 1) > 0 for index in range(size)]

def read_list_s32(stream: Stream) -> List[int]:
    return [read_s32(stream) for _ in range(read_s32(stream))]

def read_list_s16(stream: Stream) -> List[int]:
    return [read_s32(stream) for _ in range(read_u16(stream))]

def write_s8(stream: Stream, value: int) -> None:
    stream.write(pack("<b", clamp(value, -0x80, 0x7F)))

def write_u8(stream: Stream, value: int) -> None:
    stream.write(pack("<B", clamp(value, 0x00, 0xFF)))

def write_s16(stream: Stream, value: int) -> None:
    stream.write(pack("<h", clamp(value, -0x8000, 0x7FFF)))

def write_u16(stream: Stream, value: int) -> None:
    stream.write(pack("<H", clamp(value, 0x0000, 0xFFFF)))

def write_s32(stream: Stream, value: int) -> None:
    stream.write(pack("<i", clamp(value, -0x80000000, 0x7FFFFFFF)))

def write_u32(stream: Stream, value: int) -> None:
    stream.write(pack("<I", clamp(value, 0x00000000, 0xFFFFFFFF)))

def write_s64(stream: Stream, value: int) -> None:
    stream.write(pack("<q", clamp(value, -0x8000000000000000, 0x7FFFFFFFFFFFFFFF)))

def write_u64(stream: Stream, value: int) -> None:
    stream.write(pack("<Q", clamp(value, 0x0000000000000000, 0xFFFFFFFFFFFFFFFF)))

def write_boolean(stream: Stream, value: bool) -> None:
    write_u8(stream, int(bool(value)))

def write_f32(stream: Stream, value: float) -> None:
    stream.write(pack("<f", value))

def write_f64(stream: Stream, value: float) -> None:
    stream.write(pack("<d", value))

def write_gzip(stream: Stream, data: bytes) -> None:
    stream.write(compress(data))

def write_uleb128(stream: Stream, value: int) -> None:
    if value == 0:
        stream.write(b'\x00')
        return

    ret = bytearray()

    while value != 0:
        ret.append(value & 0x7F)
        value >>= 7
        if value != 0:
            ret[-1] |= 0x80

    stream.write(bytes(ret))

def write_string(stream: Stream, value: str) -> None:
    if not value:
        write_s8(stream, 0x00)
        return

    string = value.encode()
    length = len(string)

    write_s8(stream, 0x0b)
    write_uleb128(stream, length)
    stream.write(string)

def write_bool_list(stream: Stream, values: List[bool]) -> None:
    byte = 0

    for index in range(len(values)-1, -1, -1):
        if values[index]:
            byte |= 1
        if index > 0:
            byte = byte << 1

    write_u8(stream, byte)

def write_list_s32(stream: Stream, values: List[int]) -> None:
    write_s32(stream, len(values))

    for value in values:
        write_s32(stream, value)

def write_list_s16(stream: Stream, values: List[int]) -> None:
    write_u16(stream, len(values))

    for value in values:
        write_s32(stream, value)




from typing import Iterable, Tuple
from .b613 import b613
from ..constants import *
from ..types import *
from ..io import *

class b634(b613):
    """
    b634 adds the monitor & player update request packets.
    """
    version = 634

    @classmethod
    def write_monitor(cls) -> Iterable[Tuple[PacketType, bytes]]:
        yield PacketType.BanchoMonitor, b''

    @classmethod
    def read_receive_updates(cls, stream: MemoryStream) -> PresenceFilter:
        return PresenceFilter(read_s32(stream))




from .b338 import b338
from ..constants import *
from ..types import *
from ..io import *

class b340(b338):
    """
    b340 removes the "MatchChangeBeatmap" packet, and introduces the "MatchHasBeatmap" packet.
    """
    version = 340

    @classmethod
    def convert_input_packet(cls, packet: int) -> PacketType:
        if packet == 11:
            # "IrcJoin" packet
            return PacketType.BanchoIrcJoin

        if packet > 11:
            packet -= 1

        return PacketType(packet)

    @classmethod
    def convert_output_packet(cls, packet: PacketType) -> int:
        if packet is PacketType.BanchoIrcJoin:
            # "IrcJoin" packet
            return 11

        if packet >= 11:
            return packet.value + 1

        return packet.value

    @classmethod
    def read_match_change_beatmap(cls, stream: MemoryStream) -> ...:
        ...

    @classmethod
    def read_match_has_beatmap(self, stream: MemoryStream) -> None:
        pass




from typing import Iterable, Tuple
from .b20121207 import b20121207
from ..constants import *
from ..types import *
from ..io import *

class b20121211(b20121207):
    """
    b20121211 deprecates the lobby user list.
    """
    version = 20121211

    @classmethod
    def write_lobby_join(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        return []

    @classmethod
    def write_lobby_part(cls, user_id: int) -> Iterable[Tuple[PacketType, bytes]]:
        return []
